import numpy as np
import torch
import warnings
from .neurodiffeq import safe_diff as diff
from ._version_utils import deprecated_alias


class BaseCondition:
    r"""Base class for all conditions.

    A condition is a tool to `re-parameterize` the output(s) of a neural network.
    such that the re-parameterized output(s) will automatically satisfy initial conditions (ICs)
    and boundary conditions (BCs) of the PDEs/ODEs that are being solved.

    .. note::
        - The nouns *(re-)parameterization* and *condition* are used interchangeably in the documentation and library.
        - The verbs *(re-)parameterize* and *enforce* are different in that:

          - *(re)parameterize* is said of network outputs;
          - *enforce* is said of networks themselves.
    """

    def __init__(self):
        self.ith_unit = None

    def parameterize(self, output_tensor, *input_tensors):
        r"""Re-parameterizes output(s) of a network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param input_tensors: Inputs to the neural network; i.e., sampled coordinates; i.e., independent variables.
        :type input_tensors: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`

        .. note::
            This method is **abstract** for BaseCondition
        """
        raise ValueError(f"Abstract {self.__class__.__name__} cannot be parameterized")  # pragma: no cover

    def enforce(self, net, *coordinates):
        r"""Enforces this condition on a network.

        :param net: The network whose output is to be re-parameterized.
        :type net: `torch.nn.Module`
        :param coordinates: Inputs of the neural network.
        :type coordinates: `torch.Tensor`
        :return: The re-parameterized output, where the condition is automatically satisfied.
        :rtype: `torch.Tensor`
        """
        # concatenate the coordinates and pass to network
        network_output = net(torch.cat(coordinates, dim=1))
        # if `ith_unit` is set, the condition will only be enforced on the i-th output unit
        if self.ith_unit is not None:
            network_output = network_output[:, self.ith_unit].view(-1, 1)
        # parameterize the raw output and return
        return self.parameterize(network_output, *coordinates)

    def set_impose_on(self, ith_unit):
        r"""**[DEPRECATED]** When training several functions with a single, multi-output network, this method is called
        (by a `Solver` class or a `solve` function) to keep track of which output is being parameterized.

        :param ith_unit: The index of network output to be parameterized.
        :type ith_unit: int

        .. note::
            This method is deprecated and retained for backward compatibility only. Users interested in enforcing
            conditions on multi-output networks should consider using a ``neurodiffeq.conditions.EnsembleCondition``.
        """

        warnings.warn(
            f"`{self.__class__.__name__}.set_impose_on` is deprecated and will be removed in the future",
            DeprecationWarning,
        )
        self.ith_unit = ith_unit


class IrregularBoundaryCondition(BaseCondition):
    # Is there a more elegant solution?
    def in_domain(self, *coordinates):
        """Given the coordinates (numpy.ndarray), the methods returns an boolean array indicating
        whether the points lie within the domain.

        :param coordinates: Input tensors, each with shape (n_samples, 1).
        :type coordinates: `numpy.ndarray`
        :return: Whether each point lies within the domain.
        :rtype: `numpy.ndarray`

        .. note::
            - This method is meant to be used by monitors for irregular domain visualization.
        """

        # returns straight `True`-s by default; i.e., all points are considered within domain
        return np.ones_like(coordinates[0], dtype=np.bool)


class EnsembleCondition(BaseCondition):
    r"""An ensemble condition that enforces sub-conditions on individual output units of the networks.

    :param sub_conditions: Condition(s) to be ensemble'd.
    :type sub_conditions: BaseCondition
    :param force: Whether or not to force ensembl'ing even when `.enforce` is overridden in one of the sub-conditions.
    :type force: bool
    """

    def __init__(self, *sub_conditions, force=False):
        super(EnsembleCondition, self).__init__()
        for i, c in enumerate(sub_conditions):
            if c.__class__.enforce != BaseCondition.enforce:
                msg = f"{c.__class__.__name__} (index={i})'s overrides BaseCondition's `.enforce` method. " \
                      f"Ensembl'ing is likely not going to work."
                if force:
                    warnings.warn(msg)
                else:
                    raise ValueError(msg + "\nTry with `force=True` if you know what you are doing.")

        self.conditions = sub_conditions

    def parameterize(self, output_tensor, *input_tensors):
        r"""Re-parameterizes each column in output_tensor individually, using its corresponding sub-condition.
        This is useful when solving differential equations with a single, multi-output network.

        :param output_tensor:
            Output of the neural network.
            Number of units (.shape[1]) must equal number of sub-conditions.
        :type output_tensor: `torch.Tensor`
        :param input_tensors:
            Inputs to the neural network;
            i.e., sampled coordinates;
            i.e., independent variables.
        :type input_tensors: `torch.Tensor`
        :return:
            The column-wise re-parameterized network output,
            concatenated across columns so that it's still one tensor.
        :rtype: `torch.Tensor`
        """
        if output_tensor.shape[1] != len(self.conditions):
            raise ValueError(f"number of output units ({output_tensor.shape[1]}) "
                             f"differs from number of conditions ({len(self.conditions)})")
        return torch.cat([
            con.parameterize(output_tensor[:, i].view(-1, 1), *input_tensors) for i, con in enumerate(self.conditions)
        ], dim=1)


class NoCondition(BaseCondition):
    r"""A polymorphic condition where no re-parameterization will be performed.

    .. note::
        This condition is called *polymorphic* because it can be enforced on networks of arbitrary input/output sizes.
    """

    def parameterize(self, output_tensor, *input_tensors):
        r"""Performs no re-parameterization, or identity parameterization, in this case.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param input_tensors: Inputs to the neural network; i.e., sampled coordinates; i.e., independent variables.
        :type input_tensors: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """
        return output_tensor


class IVP(BaseCondition):
    r"""An initial value problem of one of the following forms:

    - Dirichlet condition: :math:`u(t_0)=u_0`.
    - Neumann condition: :math:`\displaystyle\frac{\partial u}{\partial t}\bigg|_{t = t_0} = u_0'`.

    :param t_0: The initial time.
    :type t_0: float
    :param u_0: The initial value of :math:`u`. :math:`u(t_0)=u_0`.
    :type u_0: float
    :param u_0_prime:
        The initial derivative of :math:`u` w.r.t. :math:`t`.
        :math:`\displaystyle\frac{\partial u}{\partial t}\bigg|_{t = t_0} = u_0'`.
        Defaults to None.
    :type u_0_prime: float, optional
    """

    @deprecated_alias(x_0='u_0', x_0_prime='u_0_prime')
    def __init__(self, t_0, u_0=None, u_0_prime=None):
        super().__init__()
        self.t_0, self.u_0, self.u_0_prime = t_0, u_0, u_0_prime

    def parameterize(self, output_tensor, t):
        r"""Re-parameterizes outputs such that the Dirichlet/Neumann condition is satisfied.

        - For Dirichlet condition, the re-parameterization is
          :math:`\displaystyle u(t) = u_0 + \left(1 - e^{-(t-t_0)}\right) \mathrm{ANN}(t)`
          where :math:`\mathrm{ANN}` is the neural network.
        - For Neumann condition, the re-parameterization is
          :math:`\displaystyle u(t) = u_0 + (t-t_0) u'_0 + \left(1 - e^{-(t-t_0)}\right)^2 \mathrm{ANN}(t)`
          where :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param t: Input to the neural network; i.e., sampled time-points; i.e., independent variables.
        :type t: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """
        if self.u_0_prime is None:
            return self.u_0 + (1 - torch.exp(-t + self.t_0)) * output_tensor
        else:
            return self.u_0 + (t - self.t_0) * self.u_0_prime + ((1 - torch.exp(-t + self.t_0)) ** 2) * output_tensor


class BundleIVP(BaseCondition):
    r"""An initial value problem of one of the following forms:

    - Dirichlet condition: :math:`u(t_0,\boldsymbol{\theta})=u_0`.
    - Neumann condition: :math:`\displaystyle\frac{\partial u}{\partial t}\bigg|_{t = t_0}(\boldsymbol{\theta}) = u_0'`.

    Here :math:`\boldsymbol{\theta}=(\theta_{1},\theta_{2},...,\theta_{n})\in\mathbb{R}^n`,
    where each :math:`\theta_i` represents a parameter, or a condition, of the ODE system that we want to solve.

    :param t_0: The initial time.
    :type t_0: float
    :param u_0: The initial value of :math:`u`. :math:`u(t_0,\boldsymbol{\theta})=u_0`.
    :type u_0: float
    :param u_0_prime:
        The initial derivative of :math:`u` w.r.t. :math:`t`.
        :math:`\displaystyle\frac{\partial u}{\partial t}\bigg|_{t = t_0}(\boldsymbol{\theta}) = u_0'`.
        Defaults to None.
    :type u_0_prime: float, optional
    :param bundle_conditions:
        The initial conditions that will be included in the total bundle,
        in addition to the parameters of the ODE system.
        The values asociated with their respective keys used in bundle_conditions
        (e.g bundle_conditions={'t_0': 0, 'u_0': 1, 'u_0_prime': 2}),
        must reflect the index of the tuple used in theta_min and theta_max in ``neurodiffeq.solvers.BundleSolver1D``,
        (e.g theta_min=(t_0_min, u_0_min, u_0_prime_min)).
        Defaults to {}
    :type bundle_conditions: dict{str: int, ..., str: int}
    """

    @deprecated_alias(x_0='u_0', x_0_prime='u_0_prime')
    def __init__(self, t_0=None, u_0=None, u_0_prime=None, bundle_conditions={}):
        super().__init__()
        self.t_0, self.u_0, self.u_0_prime = t_0, u_0, u_0_prime
        self.bundle_conditions = bundle_conditions

    def parameterize(self, output_tensor, t, *theta):
        r"""Re-parameterizes outputs such that the Dirichlet/Neumann condition is satisfied.

        if t_0 is not included in the bundle:

        - For Dirichlet condition, the re-parameterization is
          :math:`\displaystyle u(t,\boldsymbol{\theta}) = u_0 + \left(1 - e^{-(t-t_0)}\right)`
          :math:`\mathrm{ANN}(t,\boldsymbol{\theta})`
        - For Neumann condition, the re-parameterization is
          :math:`\displaystyle u(t,\boldsymbol{\theta}) = u_0 + (t-t_0) u'_0 + \left(1 - e^{-(t-t_0)}\right)^2`
          :math:`\mathrm{ANN}(t,\boldsymbol{\theta})`

        if t_0 is included in the bundle:

        - For Dirichlet condition, the re-parameterization is
          :math:`\displaystyle u(t,\boldsymbol{\theta}) = u_0 + \left(t - t_0\right)`
          :math:`\mathrm{ANN}(t,\boldsymbol{\theta})`
        - For Neumann condition, the re-parameterization is
          :math:`\displaystyle u(t,\boldsymbol{\theta}) = u_0 + (t-t_0) u'_0 + \left(t - t_0\right)^2`
          :math:`\mathrm{ANN}(t,\boldsymbol{\theta})`

        Where :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param t: First input to the neural network; i.e., sampled time-points; i.e., independent variables.
        :type t: `torch.Tensor`
        :param theta: Rest of the inputs to the neural network; i.e., sampled bundle-points
        :type theta: tuple[torch.Tensor, ..., torch.Tensor]
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """

        t_0, u_0, u_0_prime = self.t_0, self.u_0, self.u_0_prime
        if 'u_0' in self.bundle_conditions:
            u_0 = theta[self.bundle_conditions['u_0']]
        if 'u_0_prime' in self.bundle_conditions:
            u_0_prime = theta[self.bundle_conditions['u_0_prime']]

        if 't_0' in self.bundle_conditions:

            t_0 = theta[self.bundle_conditions['t_0']]

            if self.u_0_prime is None and 'u_0_prime' not in self.bundle_conditions:
                return u_0 + (t - t_0) * output_tensor
            else:
                return u_0 + (t - t_0) * u_0_prime + ((t - t_0) ** 2) * output_tensor
        else:
            if self.u_0_prime is None and 'u_0_prime' not in self.bundle_conditions:
                return u_0 + (1 - torch.exp(-t + t_0)) * output_tensor
            else:
                return u_0 + (t - t_0) * u_0_prime + ((1 - torch.exp(-t + t_0)) ** 2) * output_tensor


class DirichletBVP(BaseCondition):
    r"""A double-ended Dirichlet boundary condition:
    :math:`u(t_0)=u_0` and :math:`u(t_1)=u_1`.

    :param t_0: The initial time.
    :type t_0: float
    :param u_0: The initial value of :math:`u`. :math:`u(t_0)=u_0`.
    :type u_0: float
    :param t_1: The final time.
    :type t_1: float
    :param u_1: The initial value of :math:`u`. :math:`u(t_1)=u_1`.
    :type u_1: float
    """

    @deprecated_alias(x_0='u_0', x_1='u_1')
    def __init__(self, t_0, u_0, t_1, u_1):
        super().__init__()
        self.t_0, self.u_0, self.t_1, self.u_1 = t_0, u_0, t_1, u_1

    def parameterize(self, output_tensor, t):
        r"""Re-parameterizes outputs such that the Dirichlet condition is satisfied on both ends of the domain.

        The re-parameterization is
        :math:`\displaystyle u(t)=(1-\tilde{t})u_0+\tilde{t}u_1+\left(1-e^{(1-\tilde{t})\tilde{t}}\right)\mathrm{ANN}(t)`,
        where :math:`\displaystyle \tilde{t} = \frac{t-t_0}{t_1-t_0}` and :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param t: Input to the neural network; i.e., sampled time-points or another independent variable.
        :type t: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """

        t_tilde = (t - self.t_0) / (self.t_1 - self.t_0)
        return self.u_0 * (1 - t_tilde) \
               + self.u_1 * t_tilde \
               + (1 - torch.exp((1 - t_tilde) * t_tilde)) * output_tensor


class DirichletBVP2D(BaseCondition):
    r"""An Dirichlet boundary condition on the boundary of :math:`[x_0, x_1] \times [y_0, y_1]`, where

    - :math:`u(x_0, y) = f_0(y)`;
    - :math:`u(x_1, y) = f_1(y)`;
    - :math:`u(x, y_0) = g_0(x)`;
    - :math:`u(x, y_1) = g_1(x)`.

    :param x_min: The lower bound of x, the :math:`x_0`.
    :type x_min: float
    :param x_min_val: The boundary value on :math:`x = x_0`, i.e. :math:`f_0(y)`.
    :type x_min_val: callable
    :param x_max: The upper bound of x, the :math:`x_1`.
    :type x_max: float
    :param x_max_val: The boundary value on :math:`x = x_1`, i.e. :math:`f_1(y)`.
    :type x_max_val: callable
    :param y_min: The lower bound of y, the :math:`y_0`.
    :type y_min: float
    :param y_min_val: The boundary value on :math:`y = y_0`, i.e. :math:`g_0(x)`.
    :type y_min_val: callable
    :param y_max: The upper bound of y, the :math:`y_1`.
    :type y_max: float
    :param y_max_val: The boundary value on :math:`y = y_1`, i.e. :math:`g_1(x)`.
    :type y_max_val: callable
    """

    def __init__(self, x_min, x_min_val, x_max, x_max_val, y_min, y_min_val, y_max, y_max_val):
        r"""Initializer method
        """
        super().__init__()
        self.x0, self.f0 = x_min, x_min_val
        self.x1, self.f1 = x_max, x_max_val
        self.y0, self.g0 = y_min, y_min_val
        self.y1, self.g1 = y_max, y_max_val

    def parameterize(self, output_tensor, x, y):
        r"""Re-parameterizes outputs such that the Dirichlet condition is satisfied on all four sides of the domain.

        The re-parameterization is
        :math:`\displaystyle u(x,y)=A(x,y)
        +\tilde{x}\big(1-\tilde{x}\big)\tilde{y}\big(1-\tilde{y}\big)\mathrm{ANN}(x,y)`, where

        :math:`\displaystyle \begin{align*}
        A(x,y)=&\big(1-\tilde{x}\big)f_0(y)+\tilde{x}f_1(y) \\
        &+\big(1-\tilde{y}\big)\Big(g_0(x)-\big(1-\tilde{x}\big)g_0(x_0)+\tilde{x}g_0(x_1)\Big) \\
        &+\tilde{y}\Big(g_1(x)-\big(1-\tilde{x}\big)g_1(x_0)+\tilde{x}g_1(x_1)\Big)
        \end{align*}`

        :math:`\displaystyle\tilde{x}=\frac{x-x_0}{x_1-x_0}`,

        :math:`\displaystyle\tilde{y}=\frac{y-y_0}{y_1-y_0}`,

        and :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param x: :math:`x`-coordinates of inputs to the neural network; i.e., the sampled :math:`x`-coordinates.
        :type x: `torch.Tensor`
        :param y: :math:`y`-coordinates of inputs to the neural network; i.e., the sampled :math:`y`-coordinates.
        :type y: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """
        x_tilde = (x - self.x0) / (self.x1 - self.x0)
        y_tilde = (y - self.y0) / (self.y1 - self.y0)
        x0 = torch.ones_like(x_tilde[0, 0]).expand(*x_tilde.shape) * self.x0
        x1 = torch.ones_like(x_tilde[0, 0]).expand(*x_tilde.shape) * self.x1
        Axy = (1 - x_tilde) * self.f0(y) + x_tilde * self.f1(y) \
              + (1 - y_tilde) * (self.g0(x) - ((1 - x_tilde) * self.g0(x0) + x_tilde * self.g0(x1))) \
              + y_tilde * (self.g1(x) - ((1 - x_tilde) * self.g1(x0) + x_tilde * self.g1(x1)))

        return Axy + x_tilde * (1 - x_tilde) * y_tilde * (1 - y_tilde) * output_tensor


class IBVP1D(BaseCondition):
    r"""An initial & boundary condition on a 1-D range where :math:`x\in[x_0, x_1]` and time starts at :math:`t_0`.
    The conditions should have the following parts:

    - :math:`u(x,t_0)=u_0(x)`,
    - :math:`u(x_0,t)=g(t)` or :math:`u'_x(x_0,t)=p(t)`,
    - :math:`u(x_1,t)=h(t)` or :math:`u'_x(x_1,t)=q(t)`,

    where :math:`\displaystyle u'_x=\frac{\partial u}{\partial x}`.

    :param x_min: The lower bound of x, the :math:`x_0`.
    :type x_min: float
    :param x_max: The upper bound of x, the :math:`x_1`.
    :type x_max: float
    :param t_min: The initial time, the :math:`t_0`.
    :type t_min: float
    :param t_min_val: The initial condition, the :math:`u_0(x)`.
    :type t_min_val: callable
    :param x_min_val: The Dirichlet boundary condition when :math:`x = x_0`, the :math:`u(x_0, t)`, defaults to None.
    :type x_min_val: callable, optional
    :param x_min_prime: The Neumann boundary condition when :math:`x = x_0`, the :math:`u'_x(x_0, t)`, defaults to None.
    :type x_min_prime: callable, optional
    :param x_max_val: The Dirichlet boundary condition when :math:`x = x_1`, the :math:`u(x_1, t)`, defaults to None.
    :type x_max_val: callable, optional
    :param x_max_prime: The Neumann boundary condition when :math:`x = x_1`, the :math:`u'_x(x_1, t)`, defaults to None.
    :type x_max_prime: callable, optional
    :raises NotImplementedError: When unimplemented boundary conditions are configured.

    .. note::
        This condition cannot be passed to ``neurodiffeq.conditions.EnsembleCondition`` unless both boundaries uses
        Dirichlet conditions (by specifying only ``x_min_val`` and ``x_max_val``) and ``force`` is set to True in
        EnsembleCondition's constructor.
    """

    def __init__(
            self, x_min, x_max, t_min, t_min_val,
            x_min_val=None, x_min_prime=None,
            x_max_val=None, x_max_prime=None,
    ):
        super().__init__()
        n_conditions = sum(c is not None for c in [x_min_val, x_min_prime, x_max_val, x_max_prime])
        if n_conditions != 2 or (x_min_val and x_min_prime) or (x_max_val and x_max_prime):
            raise NotImplementedError('Sorry, this boundary condition is not implemented.')
        self.x_min, self.x_min_val, self.x_min_prime = x_min, x_min_val, x_min_prime
        self.x_max, self.x_max_val, self.x_max_prime = x_max, x_max_val, x_max_prime
        self.t_min, self.t_min_val = t_min, t_min_val

    def enforce(self, net, x, t):
        r"""Enforces this condition on a network with inputs `x` and `t`

        :param net: The network whose output is to be re-parameterized.
        :type net: `torch.nn.Module`
        :param x: The :math:`x`-coordinates of the samples; i.e., the spatial coordinates.
        :type x: `torch.Tensor`
        :param t: The :math:`t`-coordinates of the samples; i.e., the temporal coordinates.
        :type t: `torch.Tensor`
        :return: The re-parameterized output, where the condition is automatically satisfied.
        :rtype: `torch.Tensor`

        .. note::
            This method overrides the default method of ``neurodiffeq.conditions.BaseCondition`` .
            In general, you should avoid overriding ``enforce`` when implementing custom boundary conditions.
        """

        def ANN(x, t):
            out = net(torch.cat([x, t], dim=1))
            if self.ith_unit is not None:
                out = out[:, self.ith_unit].view(-1, 1)
            return out

        uxt = ANN(x, t)
        if self.x_min_val and self.x_max_val:
            return self.parameterize(uxt, x, t)
        elif self.x_min_val and self.x_max_prime:
            x1 = self.x_max * torch.ones_like(x, requires_grad=True)
            ux1t = ANN(x1, t)
            return self.parameterize(uxt, x, t, ux1t, x1)
        elif self.x_min_prime and self.x_max_val:
            x0 = self.x_min * torch.ones_like(x, requires_grad=True)
            ux0t = ANN(x0, t)
            return self.parameterize(uxt, x, t, ux0t, x0)
        elif self.x_min_prime and self.x_max_prime:
            x0 = self.x_min * torch.ones_like(x, requires_grad=True)
            x1 = self.x_max * torch.ones_like(x, requires_grad=True)
            ux0t = ANN(x0, t)
            ux1t = ANN(x1, t)
            return self.parameterize(uxt, x, t, ux0t, x0, ux1t, x1)
        else:
            raise NotImplementedError('Sorry, this boundary condition is not implemented.')

    def parameterize(self, u, x, t, *additional_tensors):
        r"""Re-parameterizes outputs such that the initial and boundary conditions are satisfied.

        The Initial condition is always :math:`u(x,t_0)=u_0(x)`. There are four boundary conditions that are
        currently implemented:

        - For Dirichlet-Dirichlet boundary condition :math:`u(x_0,t)=g(t)` and :math:`u(x_1,t)=h(t)`:

          The re-parameterization is
          :math:`\displaystyle u(x,t)=A(x,t)+\tilde{x}\big(1-\tilde{x}\big)\Big(1-e^{-\tilde{t}}\Big)\mathrm{ANN}(x,t)`,
          where :math:`\displaystyle A(x,t)=u_0(x)+
          \tilde{x}\big(h(t)-h(t_0)\big)+\big(1-\tilde{x}\big)\big(g(t)-g(t_0)\big)`.

        - For Dirichlet-Neumann boundary condition :math:`u(x_0,t)=g(t)` and :math:`u'_x(x_1, t)=q(t)`:

          The re-parameterization is
          :math:`\displaystyle u(x,t)=A(x,t)+\tilde{x}\Big(1-e^{-\tilde{t}}\Big)
          \Big(\mathrm{ANN}(x,t)-\big(x_1-x_0\big)\mathrm{ANN}'_x(x_1,t)-\mathrm{ANN}(x_1,t)\Big)`,
          where :math:`\displaystyle A(x,t)=u_0(x)+\big(x-x_0\big)\big(q(t)-q(t_0)\big)+\big(g(t)-g(t_0)\big)`.

        - For Neumann-Dirichlet boundary condition :math:`u'_x(x_0,t)=p(t)` and :math:`u(x_1, t)=h(t)`:

          The re-parameterization is
          :math:`\displaystyle u(x,t)=A(x,t)+\big(1-\tilde{x}\big)\Big(1-e^{-\tilde{t}}\Big)
          \Big(\mathrm{ANN}(x,t)-\big(x_1-x_0\big)\mathrm{ANN}'_x(x_0,t)-\mathrm{ANN}(x_0,t)\Big)`,
          where :math:`\displaystyle A(x,t)=u_0(x)+\big(x_1-x\big)\big(p(t)-p(t_0)\big)+\big(h(t)-h(t_0)\big)`.

        - For Neumann-Neumann boundary condition :math:`u'_x(x_0,t)=p(t)` and :math:`u'_x(x_1, t)=q(t)`

          The re-parameterization is
          :math:`\displaystyle u(x,t)=A(x,t)+\left(1-e^{-\tilde{t}}\right)
          \Big(
          \mathrm{ANN}(x,t)-\big(x-x_0\big)\mathrm{ANN}'_x(x_0,t)
          +\frac{1}{2}\tilde{x}^2\big(x_1-x_0\big)
          \big(\mathrm{ANN}'_x(x_0,t)-\mathrm{ANN}'_x(x_1,t)\big)
          \Big)`,
          where :math:`\displaystyle A(x,t)=u_0(x)
          -\frac{1}{2}\big(1-\tilde{x}\big)^2\big(x_1-x_0\big)\big(p(t)-p(t_0)\big)
          +\frac{1}{2}\tilde{x}^2\big(x_1-x_0\big)\big(q(t)-q(t_0)\big)`.

        Notations:

        - :math:`\displaystyle\tilde{t}=\frac{t-t_0}{t_1-t_0}`,
        - :math:`\displaystyle\tilde{x}=\frac{x-x_0}{x_1-x_0}`,
        - :math:`\displaystyle\mathrm{ANN}` is the neural network,
        - and :math:`\displaystyle\mathrm{ANN}'_x=\frac{\partial ANN}{\partial x}`.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param x: The :math:`x`-coordinates of the samples; i.e., the spatial coordinates.
        :type x: `torch.Tensor`
        :param t: The :math:`t`-coordinates of the samples; i.e., the temporal coordinates.
        :type t: `torch.Tensor`
        :param additional_tensors: additional tensors that will be passed by ``enforce``
        :type additional_tensors: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """

        t0 = self.t_min * torch.ones_like(t, requires_grad=True)
        x_tilde = (x - self.x_min) / (self.x_max - self.x_min)
        t_tilde = t - self.t_min

        if self.x_min_val and self.x_max_val:
            return self._parameterize_dd(u, x, t, x_tilde, t_tilde, t0)
        elif self.x_min_val and self.x_max_prime:
            return self._parameterize_dn(u, x, t, x_tilde, t_tilde, t0, *additional_tensors)
        elif self.x_min_prime and self.x_max_val:
            return self._parameterize_nd(u, x, t, x_tilde, t_tilde, t0, *additional_tensors)
        elif self.x_min_prime and self.x_max_prime:
            return self._parameterize_nn(u, x, t, x_tilde, t_tilde, t0, *additional_tensors)
        else:
            raise NotImplementedError('Sorry, this boundary condition is not implemented.')

    # When we have Dirichlet boundary conditions on both ends of the domain:
    def _parameterize_dd(self, uxt, x, t, x_tilde, t_tilde, t0):
        Axt = self.t_min_val(x) + \
              x_tilde * (self.x_max_val(t) - self.x_max_val(t0)) + \
              (1 - x_tilde) * (self.x_min_val(t) - self.x_min_val(t0))
        return Axt + x_tilde * (1 - x_tilde) * (1 - torch.exp(-t_tilde)) * uxt

    # When we have Dirichlet boundary condition on the left end of the domain
    # and Neumann boundary condition on the right end of the domain:
    def _parameterize_dn(self, uxt, x, t, x_tilde, t_tilde, t0, ux1t, x1):
        Axt = (self.x_min_val(t) - self.x_min_val(t0)) + self.t_min_val(x) + \
              x_tilde * (self.x_max - self.x_min) * (self.x_max_prime(t) - self.x_max_prime(t0))
        return Axt + x_tilde * (1 - torch.exp(-t_tilde)) * (
                uxt - (self.x_max - self.x_min) * diff(ux1t, x1) - ux1t
        )

    # When we have Neumann boundary condition on the left end of the domain
    # and Dirichlet boundary condition on the right end of the domain:
    def _parameterize_nd(self, uxt, x, t, x_tilde, t_tilde, t0, ux0t, x0):
        Axt = (self.x_max_val(t) - self.x_max_val(t0)) + self.t_min_val(x) + \
              (x_tilde - 1) * (self.x_max - self.x_min) * (self.x_min_prime(t) - self.x_min_prime(t0))
        return Axt + (1 - x_tilde) * (1 - torch.exp(-t_tilde)) * (
                uxt + (self.x_max - self.x_min) * diff(ux0t, x0) - ux0t
        )

    # When we have Neumann boundary conditions on both ends of the domain:
    def _parameterize_nn(self, uxt, x, t, x_tilde, t_tilde, t0, ux0t, x0, ux1t, x1):
        Axt = self.t_min_val(x) \
              - 0.5 * (1 - x_tilde) ** 2 * (self.x_max - self.x_min) * (self.x_min_prime(t) - self.x_min_prime(t0)) \
              + 0.5 * x_tilde ** 2 * (self.x_max - self.x_min) * (self.x_max_prime(t) - self.x_max_prime(t0))
        return Axt + (1 - torch.exp(-t_tilde)) * (
                uxt
                - x_tilde * (self.x_max - self.x_min) * diff(ux0t, x0)
                + 0.5 * x_tilde ** 2 * (self.x_max - self.x_min) * (
                        diff(ux0t, x0) - diff(ux1t, x1)
                )
        )


class DoubleEndedBVP1D(BaseCondition):
    r"""A boundary condition on a 1-D range where :math:`x\in[x_0, x_1]`.
    The conditions should have the following parts:

    - :math:`u(x_0)=u_0` or :math:`u'_x(x_0)=u'_0`,
    - :math:`u(x_1)=u_1` or :math:`u'_x(x_1)=u'_1`,

    where :math:`\displaystyle u'_x=\frac{\partial u}{\partial x}`.

    :param x_min: The lower bound of x, the :math:`x_0`.
    :type x_min: float
    :param x_max: The upper bound of x, the :math:`x_1`.
    :type x_max: float
    :param x_min_val: The Dirichlet boundary condition when :math:`x = x_0`, the :math:`u(x_0)`, defaults to None.
    :type x_min_val: callable, optional
    :param x_min_prime: The Neumann boundary condition when :math:`x = x_0`, the :math:`u'_x(x_0)`, defaults to None.
    :type x_min_prime: callable, optional
    :param x_max_val: The Dirichlet boundary condition when :math:`x = x_1`, the :math:`u(x_1)`, defaults to None.
    :type x_max_val: callable, optional
    :param x_max_prime: The Neumann boundary condition when :math:`x = x_1`, the :math:`u'_x(x_1)`, defaults to None.
    :type x_max_prime: callable, optional
    :raises NotImplementedError: When unimplemented boundary conditions are configured.

    .. note::
        This condition cannot be passed to ``neurodiffeq.conditions.EnsembleCondition`` unless both boundaries uses
        Dirichlet conditions (by specifying only ``x_min_val`` and ``x_max_val``) and ``force`` is set to True in
        EnsembleCondition's constructor.
    """

    def __init__(
            self, x_min, x_max,
            x_min_val=None, x_min_prime=None,
            x_max_val=None, x_max_prime=None,
    ):
        super().__init__()
        n_conditions = sum(c is not None for c in [x_min_val, x_min_prime, x_max_val, x_max_prime])
        if n_conditions != 2 or (x_min_val and x_min_prime) or (x_max_val and x_max_prime):
            raise NotImplementedError('Sorry, this boundary condition is not implemented.')
        self.x_min, self.x_min_val, self.x_min_prime = x_min, x_min_val, x_min_prime
        self.x_max, self.x_max_val, self.x_max_prime = x_max, x_max_val, x_max_prime

    def enforce(self, net, x):
        r"""Enforces this condition on a network with inputs `x`.

        :param net: The network whose output is to be re-parameterized.
        :type net: `torch.nn.Module`
        :param x: The :math:`x`-coordinates of the samples; i.e., the spatial coordinates.
        :type x: `torch.Tensor`
        :return: The re-parameterized output, where the condition is automatically satisfied.
        :rtype: `torch.Tensor`

        .. note::
            This method overrides the default method of ``neurodiffeq.conditions.BaseCondition`` .
            In general, you should avoid overriding ``enforce`` when implementing custom boundary conditions.
        """

        def ANN(x):
            out = net(torch.cat([x], dim=1))
            if self.ith_unit is not None:
                out = out[:, self.ith_unit].view(-1, 1)
            return out

        ux = ANN(x)
        if self.x_min_val is not None and self.x_max_val is not None:
            return self.parameterize(ux, x)
        elif self.x_min_val is not None and self.x_max_prime is not None:
            x1 = self.x_max * torch.ones_like(x, requires_grad=True)
            ux1 = ANN(x1)
            return self.parameterize(ux, x, ux1, x1)
        elif self.x_min_prime is not None and self.x_max_val is not None:
            x0 = self.x_min * torch.ones_like(x, requires_grad=True)
            ux0 = ANN(x0)
            return self.parameterize(ux, x, ux0, x0)
        elif self.x_min_prime is not None and self.x_max_prime is not None:
            x0 = self.x_min * torch.ones_like(x, requires_grad=True)
            x1 = self.x_max * torch.ones_like(x, requires_grad=True)
            ux0 = ANN(x0)
            ux1 = ANN(x1)
            return self.parameterize(ux, x, ux0, x0, ux1, x1)
        else:
            raise NotImplementedError('Sorry, this boundary condition is not implemented.')

    def parameterize(self, u, x, *additional_tensors):
        r"""Re-parameterizes outputs such that the boundary conditions are satisfied.

        There are four boundary conditions that are
        currently implemented:

        - For Dirichlet-Dirichlet boundary condition :math:`u(x_0)=u_0` and :math:`u(x_1)=u_1`:

          The re-parameterization is
          :math:`\displaystyle u(x)=A+\tilde{x}\big(1-\tilde{x}\big)\mathrm{ANN}(x)`,
          where :math:`\displaystyle A=\big(1-\tilde{x}\big)u_0+\big(\tilde{x}\big)u_1`.

        - For Dirichlet-Neumann boundary condition :math:`u(x_0)=u_0` and :math:`u'_x(x_1)=u'_1`:

          The re-parameterization is
          :math:`\displaystyle u(x)=A(x)+\tilde{x}\Big(\mathrm{ANN}(x)-\mathrm{ANN}(x_1)+x_0 - \big(x_1-x_0\big)\mathrm{ANN}'_x(x_1)\Big)`,
          where :math:`\displaystyle A(x)=\big(1-\tilde{x}\big)u_0+\frac{1}{2}\tilde{x}^2\big(x_1-x_0\big)u'_1`.

        - For Neumann-Dirichlet boundary condition :math:`u'_x(x_0)=u'_0` and :math:`u(x_1)=u_1`:

          The re-parameterization is
          :math:`\displaystyle u(x)=A(x)+\big(1-\tilde{x}\big)\Big(\mathrm{ANN}(x)-\mathrm{ANN}(x_0)+x_1 + \big(x_1-x_0\big)\mathrm{ANN}'_x(x_0)\Big)`,
          where :math:`\displaystyle A(x)=\tilde{x}u_1-\frac{1}{2}\big(1-\tilde{x}\big)^2\big(x_1-x_0\big)u'_0`.

        - For Neumann-Neumann boundary condition :math:`u'_x(x_0)=u'_0` and :math:`u'_x(x_1)=u'_1`:

          The re-parameterization is
          :math:`\displaystyle u(x)=A(x)+\frac{1}{2}\tilde{x}^2\big(\mathrm{ANN}(x)-\mathrm{ANN}(x_1)-\frac{1}{2}\mathrm{ANN}'_x(x_1)\big(x_1-x_0\big)\big),
          +\frac{1}{2}\big(1-\tilde{x}\big)^2\big(\mathrm{ANN}(x)-\mathrm{ANN}(x_0)+\frac{1}{2}\mathrm{ANN}'_x(x_0)\big(x_1-x_0\big)\big)`,
          where :math:`\displaystyle A(x)=\frac{1}{2}\tilde{x}^2\big(x_1-x_0\big)u'_1 - \frac{1}{2}\big(1-\tilde{x}\big)^2\big(x_1-x_0\big)u'_0`.


        Notations:

        - :math:`\displaystyle\tilde{x}=\frac{x-x_0}{x_1-x_0}`,
        - :math:`\displaystyle\mathrm{ANN}` is the neural network,
        - and :math:`\displaystyle\mathrm{ANN}'_x=\frac{\partial ANN}{\partial x}`.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param x: The :math:`x`-coordinates of the samples; i.e., the spatial coordinates.
        :type x: `torch.Tensor`
        :param additional_tensors: additional tensors that will be passed by ``enforce``
        :type additional_tensors: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """

        x_tilde = (x - self.x_min) / (self.x_max - self.x_min)
        if self.x_min_val is not None and self.x_max_val is not None:
            return self._parameterize_dd(u, x, x_tilde)
        elif self.x_min_val is not None and self.x_max_prime is not None:
            return self._parameterize_dn(u, x, x_tilde, *additional_tensors)
        elif self.x_min_prime is not None and self.x_max_val is not None:
            return self._parameterize_nd(u, x, x_tilde, *additional_tensors)
        elif self.x_min_prime is not None and self.x_max_prime is not None:
            return self._parameterize_nn(u, x, x_tilde, *additional_tensors)
        else:
            raise NotImplementedError('Sorry, this boundary condition is not implemented.')

    # When we have Dirichlet boundary conditions on both ends of the domain:
    def _parameterize_dd(self, ux, x, x_tilde):
        Ax = self.x_min_val * (1 - x_tilde) + self.x_max_val * (x_tilde)
        return Ax + x_tilde * (1 - x_tilde) * ux

    # When we have Dirichlet boundary condition on the left end of the domain
    # and Neumann boundary condition on the right end of the domain:
    def _parameterize_dn(self, ux, x, x_tilde, ux1, x1):
        Ax = (1 - x_tilde) * self.x_min_val + 0.5 * x_tilde ** 2 * self.x_max_prime * (self.x_max - self.x_min)
        return Ax + x_tilde * (ux - ux1 + self.x_min_val - diff(ux1, x1) * (self.x_max - self.x_min))
        # Ax = self.x_min_val + (x - self.x_min) * self.x_max_prime
        # return Ax + x_tilde * (ux - (self.x_max - self.x_min) * diff(ux1, x1) - ux1)

    # When we have Neumann boundary condition on the left end of the domain
    # and Dirichlet boundary condition on the right end of the domain:
    def _parameterize_nd(self, ux, x, x_tilde, ux0, x0):
        Ax = x_tilde * self.x_max_val - 0.5 * (1 - x_tilde) ** 2 * self.x_min_prime * (self.x_max - self.x_min)
        return Ax + (1 - x_tilde) * (ux - ux0 + self.x_max_val + diff(ux0, x0) * (self.x_max - self.x_min))
        # Ax = self.x_max_val + (x - self.x_max) * self.x_min_prime
        # return Ax + (1 - x_tilde) * (ux + (self.x_max - self.x_min) * diff(ux0, x0) - ux0)

    # When we have Neumann boundary conditions on both ends of the domain:
    def _parameterize_nn(self, ux, x, x_tilde, ux0, x0, ux1, x1):
        Ax = - 0.5 * (1 - x_tilde) ** 2 * (self.x_max - self.x_min) * self.x_min_prime + 0.5 * x_tilde ** 2 * (
                self.x_max - self.x_min) * self.x_max_prime
        return Ax + 0.5 * x_tilde ** 2 * (ux - ux1 - 0.5 * diff(ux1, x1) * (self.x_max - self.x_min)) + \
               0.5 * (1 - x_tilde) ** 2 * (ux - ux0 + 0.5 * diff(ux0, x0) * (self.x_max - self.x_min))


# TODO: reduce duplication
class DirichletBVPSpherical(BaseCondition):
    r"""The Dirichlet boundary condition for the interior and exterior boundary of the sphere,
    where the interior boundary is not necessarily a point. The conditions are:

    - :math:`u(r_0,\theta,\phi)=f(\theta,\phi)`
    - :math:`u(r_1,\theta,\phi)=g(\theta,\phi)`

    :param r_0:
        The radius of the interior boundary.
        When :math:`r_0 = 0`, the interior boundary collapses to a single point (center of the ball).
    :type r_0: float
    :param f:
        The value of :math:`u` on the interior boundary.
        :math:`u(r_0, \theta, \phi)=f(\theta, \phi)`.
    :type f: callable
    :param r_1:
        The radius of the exterior boundary.
        If set to None, `g` must also be None.
    :type r_1: float or None
    :param g:
        The value of :math:`u` on the exterior boundary.
        :math:`u(r_1, \theta, \phi)=g(\theta, \phi)`.
        If set to None, `r_1` must also be set to None.
    :type g: callable or None
    """

    def __init__(self, r_0, f, r_1=None, g=None):
        super(DirichletBVPSpherical, self).__init__()
        if (r_1 is None) ^ (g is None):
            raise ValueError(f'r_1 and g must be both/neither set to None; got r_1={r_1}, g={g}')
        self.r_0, self.r_1 = r_0, r_1
        self.f, self.g = f, g

    def parameterize(self, output_tensor, r, theta, phi):
        r"""Re-parameterizes outputs such that the Dirichlet condition is satisfied on both spherical boundaries.

        - If both inner and outer boundaries are specified
          :math:`u(r_0,\theta,\phi)=f(\theta,\phi)` and
          :math:`u(r_1,\theta,\phi)=g(\theta,\phi)`:

          The re-parameterization is
          :math:`\big(1-\tilde{r}\big)f(\theta,\phi)+\tilde{r}g(\theta,\phi)
          +\Big(1-e^{\tilde{r}(1-{\tilde{r}})}\Big)\mathrm{ANN}(r, \theta, \phi)`
          where :math:`\displaystyle\tilde{r}=\frac{r-r_0}{r_1-r_0}`;

        - If only one boundary is specified (inner or outer) :math:`u(r_0,\theta,\phi)=f(\theta,\phi)`

          The re-parameterization is
          :math:`f(\theta,\phi)+\Big(1-e^{-|r-r_0|}\Big)\mathrm{ANN}(r, \theta, \phi)`;

        where :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param r: The radii (or :math:`r`-component) of the inputs to the network.
        :type r: `torch.Tensor`
        :param theta: The co-latitudes (or :math:`\theta`-component) of the inputs to the network.
        :type theta: `torch.Tensor`
        :param phi: The longitudes (or :math:`\phi`-component) of the inputs to the network.
        :type phi: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """
        if self.r_1 is None:
            return (1 - torch.exp(-torch.abs(r - self.r_0))) * output_tensor + self.f(theta, phi)
        else:
            r_tilde = (r - self.r_0) / (self.r_1 - self.r_0)
            return self.f(theta, phi) * (1 - r_tilde) + \
                   self.g(theta, phi) * r_tilde + \
                   (1. - torch.exp((1 - r_tilde) * r_tilde)) * output_tensor


# TODO: reduce duplication
class InfDirichletBVPSpherical(BaseCondition):
    r"""Similar to ``neurodiffeq.conditions.DirichletBVPSpherical``. but with :math:`r_1\to+\infty`. Specifically,

    - :math:`\displaystyle u(r_0,\theta,\phi)=f(\theta,\phi)`,
    - :math:`\lim_{r\to+\infty}u(r,\theta,\phi)=g(\theta,\phi)`.

    :param r_0:
        The radius of the interior boundary.
        When :math:`r_0=0`, the interior boundary collapses to a single point (center of the ball).
    :type r_0: float
    :param f:
        The value of :math:`u` on the interior boundary.
        :math:`u(r_0,\theta,\phi)=f(\theta,\phi)`.
    :type f: callable
    :param g:
        The value of :math:`u` at infinity.
        :math:`\lim_{r\to+\infty}u(r,\theta,\phi)=g(\theta,\phi)`.
    :type g: callable
    :param order:
        The smallest :math:`k` such that :math:`\lim_{r\to+\infty}u(r,\theta,\phi)e^{-kr}=0`.
        Defaults to 1.
    :type order: int or float
    """

    def __init__(self, r_0, f, g, order=1):
        super(InfDirichletBVPSpherical, self).__init__()
        self.r_0 = r_0
        self.f = f
        self.g = g
        self.order = order

    def parameterize(self, output_tensor, r, theta, phi):
        r"""Re-parameterizes outputs such that the Dirichlet condition is satisfied both at :math:`r_0` and infinity.
        The re-parameterization is

        :math:`\begin{align}
        u(r,\theta,\phi)=
        &e^{-k(r-r_0)}f(\theta,\phi)\\
        &+\tanh{\big(r-r_0\big)}g(\theta,\phi)\\
        &+e^{-k(r-r_0)}\tanh{\big(r-r_0\big)}\mathrm{ANN}(r,\theta,\phi)
        \end{align}`,

        where :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param r: The radii (or :math:`r`-component) of the inputs to the network.
        :type r: `torch.Tensor`
        :param theta: The co-latitudes (or :math:`\theta`-component) of the inputs to the network.
        :type theta: `torch.Tensor`
        :param phi: The longitudes (or :math:`\phi`-component) of the inputs to the network.
        :type phi: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """

        dr = r - self.r_0
        return self.f(theta, phi) * torch.exp(-self.order * dr) + \
               self.g(theta, phi) * torch.tanh(dr) + \
               torch.exp(-self.order * dr) * torch.tanh(dr) * output_tensor


# TODO: reduce duplication
class DirichletBVPSphericalBasis(BaseCondition):
    r"""Similar to ``neurodiffeq.conditions.DirichletBVPSpherical``.
    The only difference is this condition is enforced on a neural net that only takes in :math:`r`
    and returns the spherical harmonic coefficients R(r).
    We constrain the coefficients :math:`R_k(r)` in :math:`u(r,\theta,\phi)=\sum_{k}R_k(r)Y_k(\theta,\phi)`,
    where :math:`\big\{Y_k(\theta,\phi)\big\}_{k=1}^{K}` can be **any spherical function basis**.
    A recommended choice is the real spherical harmonics :math:`Y_l^m(\theta,\phi)`,
    where :math:`l` is the degree of the spherical harmonics and :math:`m` is the order of the spherical harmonics.

    The boundary conditions are: :math:`\mathbf{R}(r_0)=\mathbf{R}_0` and :math:`\mathbf{R}(r_1)=\mathbf{R}_1`,
    where :math:`\mathbf{R}` is a vector whose components are :math:`\big\{R_k\big\}_{k=1}^{K}`

    :param r_0:
        The radius of the interior boundary.
        When r_0 = 0, the interior boundary is collapsed to a single point (center of the ball).
    :type r_0: float
    :param R_0:
        The value of harmonic coefficients :math:`\mathbf{R}` on the interior boundary.
        :math:`\mathbf{R}(r_0)=\mathbf{R}_0`.
    :type R_0: `torch.Tensor`
    :param r_1:
        The radius of the exterior boundary.
        If set to None, `R_1` must also be None
    :type r_1: float or None
    :param R_1:
        The value of harmonic coefficients :math:`\mathbf{R}` on the exterior boundary.
        :math:`\mathbf{R}(r_1)=\mathbf{R}_1`.
    :type R_1: `torch.Tensor`
    :param max_degree:
        **[DEPRECATED]** Highest degree when using spherical harmonics.
    :type max_degree: int
    """

    def __init__(self, r_0, R_0, r_1=None, R_1=None, max_degree=None):
        super(DirichletBVPSphericalBasis, self).__init__()
        if (r_1 is None) ^ (R_1 is None):
            raise ValueError(f'r_1 and R_1 must be both/neither set to None; got r_1={r_1}, R_1={R_1}')
        self.r_0, self.r_1 = r_0, r_1
        self.R_0, self.R_1 = R_0, R_1

    def parameterize(self, output_tensor, r):
        r"""Re-parameterizes outputs such that the Dirichlet condition is satisfied on both spherical boundaries.

        - If both inner and outer boundaries are specified
          :math:`\mathbf{R}(r_0,\theta,\phi)=\mathbf{R}_0` and
          :math:`\mathbf{R}(r_1,\theta,\phi)=\mathbf{R}_1`.

          The re-parameterization is
          :math:`\big(1-\tilde{r}\big)\mathbf{R}_0+\tilde{r}\mathbf{R}_1
          +\Big(1-e^{\tilde{r}(1-{\tilde{r}})}\Big)\mathrm{ANN}(r)`
          where :math:`\displaystyle\tilde{r}=\frac{r-r_0}{r_1-r_0}`;

        - If only one boundary is specified (inner or outer) :math:`\mathbf{R}(r_0,\theta,\phi)=\mathbf{R_0}`

          The re-parameterization is
          :math:`\mathbf{R}_0+\Big(1-e^{-|r-r_0|}\Big)\mathrm{ANN}(r)`;

        where :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param r: The radii (or :math:`r`-component) of the inputs to the network.
        :type r: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """
        if self.r_1 is None:
            ret = (1 - torch.exp(-r + self.r_0)) * output_tensor + self.R_0
        else:
            r_tilde = (r - self.r_0) / (self.r_1 - self.r_0)
            ret = self.R_0 * (1 - r_tilde) + \
                  self.R_1 * r_tilde + \
                  (1. - torch.exp((1 - r_tilde) * r_tilde)) * output_tensor
        return ret


# TODO: reduce duplication
class InfDirichletBVPSphericalBasis(BaseCondition):
    r"""Similar to ``neurodiffeq.conditions.InfDirichletBVPSpherical``.
    The only difference is this condition is enforced on a neural net that only takes in :math:`r`
    and returns the spherical harmonic coefficients R(r).
    We constrain the coefficients :math:`R_k(r)` in :math:`u(r,\theta,\phi)=\sum_{k}R_k(r)Y_k(\theta,\phi)`,
    where :math:`\big\{Y_k(\theta,\phi)\big\}_{k=1}^{K}` can be **any spherical function basis**.
    A recommended choice is the real spherical harmonics :math:`Y_l^m(\theta,\phi)`,
    where :math:`l` is the degree of the spherical harmonics and :math:`m` is the order of the spherical harmonics.

    The boundary conditions are:
    :math:`\mathbf{R}(r_0)=\mathbf{R}_0` and
    :math:`\lim_{r_0\to+\infty}\mathbf{R}(r)=\mathbf{R}_1`,
    where :math:`\mathbf{R}` is a vector whose components are :math:`\big\{R_k\big\}_{k=1}^{K}`.

    :param r_0:
        The radius of the interior boundary.
        When r_0 = 0, the interior boundary is collapsed to a single point (center of the ball).
    :type r_0: float
    :param R_0:
        The value of harmonic coefficients :math:`R` on the interior boundary.
        :math:`R(r_0)=R_0`.
    :type R_0: `torch.Tensor`
    :param R_inf:
        The value of harmonic coefficients :math:`R` at infinity.
        :math:`\lim_{r\to+\infty}R(r)=R_\infty`.
    :type R_inf: `torch.Tensor`
    :param order:
        The smallest :math:`k` that guarantees :math:`\lim_{r \to +\infty} R(r) e^{-k r} = \bf 0`.
        Defaults to 1.
    :type order: int or float
    :param max_degree:
        **[DEPRECATED]** Highest degree when using spherical harmonics.
    :type max_degree: int
    """

    def __init__(self, r_0, R_0, R_inf, order=1, max_degree=None):
        super(InfDirichletBVPSphericalBasis, self).__init__()
        self.r_0 = r_0
        self.R_0 = R_0
        self.R_inf = R_inf
        self.order = order

    def parameterize(self, output_tensor, r):
        r"""Re-parameterizes outputs such that the Dirichlet condition is satisfied at both :math:`r_0` and infinity.

        The re-parameterization is

        :math:`\begin{align}
        u(r,\theta,\phi)=
        &e^{-k(r-r_0)}\mathbf{R}_0\\
        &+\tanh{\big(r-r_0\big)}\mathbf{R}_1\\
        &+e^{-k(r-r_0)}\tanh{\big(r-r_0\big)}\mathrm{ANN}(r)
        \end{align}`,

        where :math:`\mathrm{ANN}` is the neural network.

        :param output_tensor: Output of the neural network.
        :type output_tensor: `torch.Tensor`
        :param r: The radii (or :math:`r`-component) of the inputs to the network.
        :type r: `torch.Tensor`
        :return: The re-parameterized output of the network.
        :rtype: `torch.Tensor`
        """
        dr = r - self.r_0
        return self.R_0 * torch.exp(-self.order * dr) + \
               self.R_inf * torch.tanh(dr) + \
               torch.exp(-self.order * dr) * torch.tanh(dr) * output_tensor



import pytest
import torch
import numpy as np
import torch.nn as nn
from numpy import isclose
from neurodiffeq.function_basis import LegendrePolynomial
from neurodiffeq.function_basis import LegendreBasis
from neurodiffeq.function_basis import ZonalSphericalHarmonics
from neurodiffeq.function_basis import ZonalSphericalHarmonicsLaplacian
from neurodiffeq.neurodiffeq import safe_diff as diff
from scipy.special import legendre  # legendre polynomials
from scipy.special import sph_harm  # spherical harmonics


@pytest.fixture
def n_samples():
    return 50


@pytest.fixture
def shape(n_samples):
    return (n_samples, 1)


@pytest.fixture
def max_degree():
    return 20


def test_legendre_polynomials(shape, max_degree):
    x1 = np.random.rand(*shape)
    x2 = torch.tensor(x1, requires_grad=True)

    for d in range(max_degree):
        p1 = legendre(d)(x1)
        p2 = LegendrePolynomial(d)(x2)
        assert p2.requires_grad, f"output seems detached from the graph"
        p2 = p2.detach().cpu().numpy()
        assert isclose(p2, p1).all(), f"p1 = {p1}, p2 = {p2}, delta = {p1 - p2}, max_delta = {np.max(abs(p1 - p2))}"


def test_legendre_basis(shape, max_degree):
    x1 = np.random.rand(*shape)
    x2 = torch.tensor(x1, requires_grad=True)

    y1 = np.concatenate(
        [legendre(d)(x1) for d in range(max_degree + 1)],
        axis=1,
    )
    net = LegendreBasis(max_degree=max_degree)
    y2 = net(x2)
    assert y2.requires_grad, f"output seems detached from the graph"

    y2 = y2.detach().cpu().numpy()
    assert isclose(y2, y1).all(), f"y1 = {y1}, y2 = {y2}, delta = {y1 - y2}, max_delta = {np.max(abs(y1 - y2))}"


def test_zero_order_spherical_harmonics(shape, max_degree):
    # note that in scipy, theta is azimuthal angle (0, 2 pi) while phi is polar angle (0, pi)
    thetas1 = np.random.rand(*shape) * np.pi * 2
    phis1 = np.random.rand(*shape) * np.pi
    # in neurodiffeq, theta and phi should be exchanged
    thetas2 = torch.tensor(phis1, requires_grad=True)
    phis2 = torch.tensor(thetas1, requires_grad=True)
    order = 0

    y1 = np.concatenate(
        [sph_harm(order, degree, thetas1, phis1) for degree in range(max_degree + 1)],
        axis=1,
    )
    assert (np.imag(y1) == 0).all(), f"y1 has non-zero imaginary part: {y1}"
    y1 = np.real(y1)

    net = ZonalSphericalHarmonics(max_degree)
    y2 = net(thetas2, phis2)
    assert y2.requires_grad, f"output seems detached from the graph"

    y2 = y2.detach().cpu().numpy()
    assert isclose(y2, y1, atol=1e-5, rtol=1e-3).all(), \
        f"y1 = {y1}, y2 = {y2}, delta = {y1 - y2}, max_delta = {np.max(abs(y1 - y2))}"


def test_zero_order_spherical_harmonics_laplacian(shape, max_degree):
    # Somehow, if changing default dtype to float32, the test fails by a large margin
    N_FLOAT = np.float64
    T_FLOAT = torch.float64

    THETA_EPS = 0.1
    r_values = np.random.rand(*shape).astype(N_FLOAT) + 1.1
    theta_values = np.random.uniform(THETA_EPS, np.pi - THETA_EPS, size=shape).astype(N_FLOAT)
    phi_values = np.random.rand(*shape).astype(N_FLOAT) * np.pi * 2

    net = nn.Sequential(
        nn.Linear(1, 10),
        nn.Tanh(),
        nn.Linear(10, max_degree + 1),
    ).to(T_FLOAT)

    harmonics = ZonalSphericalHarmonics(max_degree=max_degree)

    r1 = torch.tensor(r_values, requires_grad=True)
    theta1 = torch.tensor(theta_values, requires_grad=True)
    phi1 = torch.tensor(phi_values, requires_grad=True)
    coeffs1 = net(r1)

    us = torch.sum(coeffs1 * harmonics(theta1, phi1), dim=1, keepdim=True)

    def laplacian1(u, r, theta, phi):
        r_lap = diff(u * r, r, order=2) / r
        theta_lap = diff(diff(u, theta) * torch.sin(theta), theta) / (r ** 2) / torch.sin(theta)
        phi_lap = diff(u, phi, order=2) / (r ** 2) / torch.sin(theta) ** 2
        return r_lap + theta_lap + phi_lap

    lap1 = laplacian1(us, r1, theta1, phi1)
    assert lap1.requires_grad, "lap1 seems detached from graph"

    r2 = torch.tensor(r_values, requires_grad=True)
    theta2 = torch.tensor(theta_values, requires_grad=True)
    phi2 = torch.tensor(phi_values, requires_grad=True)
    coeffs2 = net(r2)

    laplacian2 = ZonalSphericalHarmonicsLaplacian(max_degree=max_degree)
    lap2 = laplacian2(coeffs2, r2, theta2, phi2)
    assert lap2.requires_grad, "lap2 seems detached from graph"

    assert torch.isclose(lap2, lap1, rtol=1e-3, atol=1e-5).all(), \
        f"lap1 = {lap1}\nlap2 = {lap2}\ndelta = {lap1 - lap2}\nmax_delta = {(lap1 - lap2).abs().max().item()}"



import torch
import warnings
from .neurodiffeq import diff
from .utils import set_tensor_type as _set_tensor_type  # Don't export this function

from . import conditions
from . import function_basis
from . import generators
from . import networks
from . import neurodiffeq
from . import operators
from . import pde
from . import ode
from . import pde_spherical
from . import temporal
from . import solvers
from . import callbacks
from . import monitors
from . import utils

# Set default float type to 64 bits
_set_tensor_type(float_bits=64)

# Turn on future warning by default
warnings.simplefilter('always', FutureWarning)



"""This module contains atomic generator classes and useful tools to construct complex generators out of atomic ones
"""
import torch
import numpy as np
from typing import List


class BaseGenerator:
    """Base class for all generators; Children classes must implement a `.get_examples` method and a `.size` field.
    """

    def __init__(self):
        self.size = None

    def get_examples(self) -> List[torch.Tensor]:
        pass  # pragma: no cover

    @staticmethod
    def check_generator(obj):
        if not isinstance(obj, BaseGenerator):
            raise ValueError(f"{obj} is not a generator")

    def __add__(self, other):
        self.check_generator(other)
        return ConcatGenerator(self, other)

    def __mul__(self, other):
        self.check_generator(other)
        return EnsembleGenerator(self, other)

    def __xor__(self, other):
        self.check_generator(other)
        return MeshGenerator(self, other)

    def _internal_vars(self) -> dict:
        return dict(size=self.size)

    @staticmethod
    def _obj_repr(obj) -> str:
        if isinstance(obj, tuple):
            return '(' + ', '.join(BaseGenerator._obj_repr(item) for item in obj) + ')'
        if isinstance(obj, list):
            return '[' + ', '.join(BaseGenerator._obj_repr(item) for item in obj) + ']'
        if isinstance(obj, set):
            return '{' + ', '.join(BaseGenerator._obj_repr(item) for item in obj) + '}'
        if isinstance(obj, dict):
            return '{' + ', '.join(
                BaseGenerator._obj_repr(k) + ': ' + BaseGenerator._obj_repr(obj[k])
                for k in obj
            ) + '}'

        if isinstance(obj, torch.Tensor):
            return f'tensor(shape={tuple(obj.shape)})'
        if isinstance(obj, np.ndarray):
            return f'ndarray(shape={tuple(obj.shape)})'
        return repr(obj)

    def __repr__(self):
        d = self._internal_vars()
        keys = ', '.join(f'{k}={self._obj_repr(d[k])}' for k in d)
        return f'{self.__class__.__name__}({keys})'


class Generator1D(BaseGenerator):
    """An example generator for generating 1-D training points.

    :param size: The number of points to generate each time `get_examples` is called.
    :type size: int
    :param t_min: The lower bound of the 1-D points generated, defaults to 0.0.
    :type t_min: float, optional
    :param t_max: The upper boound of the 1-D points generated, defaults to 1.0.
    :type t_max: float, optional
    :param method:
        The distribution of the 1-D points generated.

        - If set to 'uniform',
          the points will be drew from a uniform distribution Unif(t_min, t_max).
        - If set to 'equally-spaced',
          the points will be fixed to a set of linearly-spaced points that go from t_min to t_max.
        - If set to 'equally-spaced-noisy', a normal noise will be added to the previously mentioned set of points.
        - If set to 'log-spaced', the points will be fixed to a set of log-spaced points that go from t_min to t_max.
        - If set to 'log-spaced-noisy', a normal noise will be added to the previously mentioned set of points,

        defaults to 'uniform'.
    :type method: str, optional
    :raises ValueError: When provided with an unknown method.
    """

    def __init__(self, size, t_min=0.0, t_max=1.0, method='uniform', noise_std=None):
        r"""Initializer method

        .. note::
            A instance method `get_examples` is dynamically created to generate 1-D training points.
            It will be called by the function `solve` and `solve_system`.
        """
        super(Generator1D, self).__init__()
        self.size = size
        self.t_min, self.t_max = t_min, t_max
        self.method = method
        if noise_std:
            self.noise_std = noise_std
        else:
            self.noise_std = ((t_max - t_min) / size) / 4.0
        if method == 'uniform':
            self.examples = torch.zeros(self.size, requires_grad=True)
            self.getter = lambda: self.examples + torch.rand(self.size) * (self.t_max - self.t_min) + self.t_min
        elif method == 'equally-spaced':
            self.examples = torch.linspace(self.t_min, self.t_max, self.size, requires_grad=True)
            self.getter = lambda: self.examples
        elif method == 'equally-spaced-noisy':
            self.examples = torch.linspace(self.t_min, self.t_max, self.size, requires_grad=True)
            self.getter = lambda: torch.normal(mean=self.examples, std=self.noise_std)
        elif method == 'log-spaced':
            self.examples = torch.logspace(self.t_min, self.t_max, self.size, requires_grad=True)
            self.getter = lambda: self.examples
        elif method == 'log-spaced-noisy':
            self.examples = torch.logspace(self.t_min, self.t_max, self.size, requires_grad=True)
            self.getter = lambda: torch.normal(mean=self.examples, std=self.noise_std)
        else:
            raise ValueError(f'Unknown method: {method}')

    def get_examples(self):
        return self.getter()

    def _internal_vars(self):
        d = super(Generator1D, self)._internal_vars()
        d.update(dict(
            t_min=self.t_min,
            t_max=self.t_max,
            method=self.method,
            noise_std=self.noise_std,
        ))
        return d


class Generator2D(BaseGenerator):
    r"""An example generator for generating 2-D training points.

        :param grid:
            The discretization of the 2 dimensions.
            If we want to generate points on a :math:`m \times n` grid, then `grid` is `(m, n)`.
            Defaults to `(10, 10)`.
        :type grid: tuple[int, int], optional
        :param xy_min:
            The lower bound of 2 dimensions.
            If we only care about :math:`x \geq x_0` and :math:`y \geq y_0`, then `xy_min` is `(x_0, y_0)`.
            Defaults to `(0.0, 0.0)`.
        :type xy_min: tuple[float, float], optional
        :param xy_max:
            The upper boound of 2 dimensions.
            If we only care about :math:`x \leq x_1` and :math:`y \leq y_1`, then `xy_min` is `(x_1, y_1)`.
            Defaults to `(1.0, 1.0)`.
        :type xy_max: tuple[float, float], optional
        :param method:
            The distribution of the 2-D points generated.

            - If set to 'equally-spaced', the points will be fixed to the grid specified.
            - If set to 'equally-spaced-noisy', a normal noise will be added to the previously mentioned set of points.

            Defaults to 'equally-spaced-noisy'.
        :type method: str, optional
        :param xy_noise_std:
            The standard deviation of the noise on the x and y dimension.
            If not specified, the default value will be
            (``grid step size on x dimension`` / 4, ``grid step size on y dimension`` / 4).
        :type xy_noise_std: tuple[int, int], optional, defaults to None
        :raises ValueError: When provided with an unknown method.
    """

    def __init__(self, grid=(10, 10), xy_min=(0.0, 0.0), xy_max=(1.0, 1.0), method='equally-spaced-noisy',
                 xy_noise_std=None):
        r"""Initializer method

        .. note::
            A instance method `get_examples` is dynamically created to generate 2-D training points.
            It will be called by the function `solve2D`.
        """
        super(Generator2D, self).__init__()
        self.grid = grid
        self.size = grid[0] * grid[1]
        self.xy_min = xy_min
        self.xy_max = xy_max
        self.method = method
        self.xy_noise_std = xy_noise_std

        if method == 'equally-spaced':
            x = torch.linspace(xy_min[0], xy_max[0], grid[0], requires_grad=True)
            y = torch.linspace(xy_min[1], xy_max[1], grid[1], requires_grad=True)
            # noinspection PyTypeChecker
            grid_x, grid_y = torch.meshgrid(x, y)
            self.grid_x, self.grid_y = grid_x.flatten(), grid_y.flatten()

            self.getter = lambda: (self.grid_x, self.grid_y)

        elif method == 'equally-spaced-noisy':
            x = torch.linspace(xy_min[0], xy_max[0], grid[0], requires_grad=True)
            y = torch.linspace(xy_min[1], xy_max[1], grid[1], requires_grad=True)
            # noinspection PyTypeChecker
            grid_x, grid_y = torch.meshgrid(x, y)
            self.grid_x, self.grid_y = grid_x.flatten(), grid_y.flatten()

            if xy_noise_std:
                self.noise_xstd, self.noise_ystd = xy_noise_std
            else:
                self.noise_xstd = ((xy_max[0] - xy_min[0]) / grid[0]) / 4.0
                self.noise_ystd = ((xy_max[1] - xy_min[1]) / grid[1]) / 4.0
            self.getter = lambda: (
                torch.normal(mean=self.grid_x, std=self.noise_xstd),
                torch.normal(mean=self.grid_y, std=self.noise_ystd)
            )
        else:
            raise ValueError(f'Unknown method: {method}')

    def get_examples(self):
        return self.getter()

    def _internal_vars(self) -> dict:
        d = super(Generator2D, self)._internal_vars()
        d.update(dict(
            grid=self.grid,
            xy_min=self.xy_min,
            xy_max=self.xy_max,
            method=self.method,
            xy_noise_std=self.xy_noise_std,
        ))
        return d


class Generator3D(BaseGenerator):
    r"""An example generator for generating 3-D training points. NOT TO BE CONFUSED with `GeneratorSpherical`

        :param grid:
            The discretization of the 3 dimensions.
            If we want to generate points on a :math:`m \times n \times k` grid,
            then `grid` is `(m, n, k)`, defaults to `(10, 10, 10)`.
        :type grid: tuple[int, int, int], optional
        :param xyz_min:
            The lower bound of 3 dimensions.
            If we only care about :math:`x \geq x_0`, :math:`y \geq y_0`,
            and :math:`z \geq z_0` then `xyz_min` is :math:`(x_0, y_0, z_0)`.
            Defaults to `(0.0, 0.0, 0.0)`.
        :type xyz_min: tuple[float, float, float], optional
        :param xyz_max:
            The upper bound of 3 dimensions.
            If we only care about :math:`x \leq x_1`, :math:`y \leq y_1`, i
            and :math:`z \leq z_1` then `xyz_max` is :math:`(x_1, y_1, z_1)`.
            Defaults to `(1.0, 1.0, 1.0)`.
        :type xyz_max: tuple[float, float, float], optional
        :param method:
            The distribution of the 3-D points generated.

            - If set to 'equally-spaced', the points will be fixed to the grid specified.
            - If set to 'equally-spaced-noisy', a normal noise will be added to the previously mentioned set of points.

            Defaults to 'equally-spaced-noisy'.
        :type method: str, optional
        :raises ValueError: When provided with an unknown method.
    """

    def __init__(self, grid=(10, 10, 10), xyz_min=(0.0, 0.0, 0.0), xyz_max=(1.0, 1.0, 1.0),
                 method='equally-spaced-noisy'):
        r"""Initializer method

        .. note::
            A instance method `get_examples` is dynamically created to generate 2-D training points.
            It will be called by the function `solve2D`.
        """
        super(Generator3D, self).__init__()
        self.size = grid[0] * grid[1] * grid[2]
        self.grid = grid
        self.xyz_min = xyz_min
        self.xyz_max = xyz_max
        self.method = method

        x = torch.linspace(xyz_min[0], xyz_max[0], grid[0], requires_grad=True)
        y = torch.linspace(xyz_min[1], xyz_max[1], grid[1], requires_grad=True)
        z = torch.linspace(xyz_min[2], xyz_max[2], grid[2], requires_grad=True)
        # noinspection PyTypeChecker
        grid_x, grid_y, grid_z = torch.meshgrid(x, y, z)
        self.grid_x, self.grid_y, self.grid_z = grid_x.flatten(), grid_y.flatten(), grid_z.flatten()

        if method == 'equally-spaced':
            self.getter = lambda: (self.grid_x, self.grid_y, self.grid_z)
        elif method == 'equally-spaced-noisy':
            self.noise_xmean = torch.zeros(self.size)
            self.noise_ymean = torch.zeros(self.size)
            self.noise_zmean = torch.zeros(self.size)
            self.noise_xstd = torch.ones(self.size) * ((xyz_max[0] - xyz_min[0]) / grid[0]) / 4.0
            self.noise_ystd = torch.ones(self.size) * ((xyz_max[1] - xyz_min[1]) / grid[1]) / 4.0
            self.noise_zstd = torch.ones(self.size) * ((xyz_max[2] - xyz_min[2]) / grid[2]) / 4.0
            self.getter = lambda: (
                self.grid_x + torch.normal(mean=self.noise_xmean, std=self.noise_xstd),
                self.grid_y + torch.normal(mean=self.noise_ymean, std=self.noise_ystd),
                self.grid_z + torch.normal(mean=self.noise_zmean, std=self.noise_zstd),
            )
        else:
            raise ValueError(f'Unknown method: {method}')

    def get_examples(self):
        return self.getter()

    def _internal_vars(self) -> dict:
        d = super(Generator3D, self)._internal_vars()
        d.update(dict(
            grid=self.grid,
            xyz_min=self.xyz_min,
            xyz_max=self.xyz_max,
            method=self.method,
        ))
        return d


class GeneratorND(BaseGenerator):
    r"""An example generator for generating N-D training points.

        :param grid:
            The discretization of the N dimensions.
            If we want to generate points on a :math:`n_1 \times n_2 \times ... \times n_N` grid,
            then `grid` is `(n_1, n_2, ... , n_N)`.
            Defaults to `(10, 10)`.
        :type grid: tuple[int, int, ... , int], or it can be int if N=1, optional
        :param r_min:
            The lower bound of N dimensions.
            If we only care about :math:`r_1 \geq r_1^{min}`, :math:`r_2 \geq r_2^{min}`, ... ,
            and :math:`r_N \geq r_N^{min}` then `r_min` is `(r_1_min, r_2_min, ... , r_N_min)`.
            Defaults to `(0.0, 0.0)`.
        :type r_min: tuple[float, ... , float], or it can be float if N=1, optional
        :param r_max:
            The upper boound of N dimensions.
            If we only care about :math:`r_1 \leq r_1^{max}`, :math:`r_2 \leq r_2^{max}`, ... ,
            and :math:`r_N \leq r_N^{max}` then `r_max` is `(r_1_max, r_2_max, ... , r_N_max)`.
            Defaults to `(1.0, 1.0)`.
        :type r_max: tuple[float, ... , float], or it can be float if N=1, optional
        :param methods:
            The a list of the distributions of each of the 1-D points generated that make the total N-D points.

            - If set to 'uniform',
              the points will be drew from a uniform distribution Unif(r_min[i], r_max[i]).
            - If set to 'equally-spaced',
              the points will be fixed to a set of linearly-spaced points that go from r_min[i] to r_max[i].
            - If set to 'log-spaced',
              the points will be fixed to a set of log-spaced points that go from r_min[i] to r_max[i].
            - If set to 'exp-spaced',
              the points will be fixed to a set of exp-spaced points that go from r_min[i] to r_max[i].

            Defaults to ['equally-spaced', 'equally-spaced'].
        :type methods: list[str, str, ... , str], or it can be str if N=1, optional
        :param noisy:
            if set to True a normal noise will be added to all of the N sets of points that make the generator.
            Defaults to True.
        :type noisy: bool
        :raises ValueError: When provided with unknown methods.
    """

    def __init__(self, grid=(10, 10), r_min=(0.0, 0.0), r_max=(1.0, 1.0),
                 methods=['equally-spaced', 'equally-spaced'], noisy=True, r_noise_std=None,
                 **kwargs):

        super(GeneratorND, self).__init__()
        self.size = np.prod(grid)
        self.grid = grid
        self.r_min = r_min
        self.r_max = r_max
        self.methods = methods
        self.noisy = noisy
        self.r_noise_std = r_noise_std

        if isinstance(methods, str):
            methods = [methods]

        if isinstance(grid, int):
            grid = (grid,)

        if isinstance(r_min, float) or isinstance(r_min, int):
            r_min = (r_min,)

        if isinstance(r_max, float) or isinstance(r_max, int):
            r_max = (r_max,)

        if isinstance(r_noise_std, float) or isinstance(r_noise_std, int):
            r_noise_std = (r_noise_std,)

        N = len(grid)

        cut = kwargs.pop('cut', tuple((None, None) for i in range(N)))

        base = kwargs.pop('base', tuple(10 for i in range(N)))

        abs_value = kwargs.pop('abs_value', False)

        if kwargs:
            raise ValueError(f'Unknown keyword argument(s): {list(kwargs.keys())}')

        if isinstance(base, float) or isinstance(base, int):
            base = (base,)

        if isinstance(cut[0], float) or isinstance(cut[0], int) or cut[0] is None:
            cut = (cut,)

        r = []
        r_noise_std_list = []
        for i in range(N):

            method = methods[i]

            if r_noise_std:
                noise_rstd = r_noise_std[i]
            else:
                noise_rstd = ((r_max[i] - r_min[i]) / grid[i]) / 4.0

            if method == 'equally-spaced':
                x = torch.linspace(r_min[i], r_max[i], grid[i], requires_grad=True)
                noise_rstd_tensor = noise_rstd * torch.ones(x.size())

            elif method == 'uniform':
                x = torch.zeros(grid[i], requires_grad=True)
                x = x + torch.rand(grid[i]) * (r_max[i] - r_min[i]) + r_min[i]
                noise_rstd_tensor = torch.zeros(x.size())

            elif method == 'log-spaced':
                r_min_log = np.log10(r_min[i])
                r_max_log = np.log10(r_max[i])
                x = torch.logspace(r_min_log, r_max_log, grid[i], requires_grad=True)
                noise_rstd_tensor = (noise_rstd * torch.logspace(r_min_log, r_max_log, grid[i]))

            elif method == 'exp-spaced':
                r_min_exp = base[i] ** r_min[i]
                r_max_exp = base[i] ** r_max[i]
                x = torch.linspace(r_min_exp, r_max_exp, grid[i], requires_grad=True)
                x = (torch.log(x)/np.log(base[i])).clone().detach().requires_grad_(True)
                noise_rstd_tensor = (noise_rstd * x).clone().detach()

            else:
                raise ValueError(f'Unknown method: {method}')

            x = x[cut[i][0]:cut[i][1]]
            noise_rstd_tensor = noise_rstd_tensor[cut[i][0]:cut[i][1]]
            r.append(x)
            r_noise_std_list.append(noise_rstd_tensor)

        grid_r = torch.meshgrid(r)
        grid_std = torch.meshgrid(r_noise_std_list)
        self.grid_r = [grid_r[j].flatten() for j in range(N)]
        self.grid_std = [grid_std[j].flatten() for j in range(N)]
        if noisy:
            if abs_value:
                self.getter = lambda: tuple(torch.abs(torch.normal(self.grid_r[n], self.grid_std[n]))for n in range(N))
            else:
                self.getter = lambda: tuple(torch.normal(self.grid_r[n], self.grid_std[n]) for n in range(N))
        else:
            self.getter = lambda: tuple(self.grid_r[n] for n in range(N))

    def get_examples(self):
        return self.getter()

    def _internal_vars(self) -> dict:
        d = super(GeneratorND, self)._internal_vars()
        d.update(dict(
            grid=self.grid,
            r_min=self.r_min,
            r_max=self.r_max,
            methods=self.methods,
            noisy=self.noisy,
            r_noise_std=self.r_noise_std
        ))
        return d


class GeneratorSpherical(BaseGenerator):
    r"""A generator for generating points in spherical coordinates.

    :param size: Number of points in 3-D sphere.
    :type size: int
    :param r_min: Radius of the interior boundary.
    :type r_min: float, optional
    :param r_max: Radius of the exterior boundary.
    :type r_max: float, optional
    :param method:
        The distribution of the 3-D points generated.

        - If set to 'equally-radius-noisy', radius of the points will be drawn
          from a uniform distribution :math:`r \sim U[r_{min}, r_{max}]`.
        - If set to 'equally-spaced-noisy', squared radius of the points will be drawn
          from a uniform distribution :math:`r^2 \sim U[r_{min}^2, r_{max}^2]`

        Defaults to 'equally-spaced-noisy'.

    :type method: str, optional

    .. note::
        Not to be confused with ``Generator3D``.
    """

    # noinspection PyMissingConstructor
    def __init__(self, size, r_min=0., r_max=1., method='equally-spaced-noisy'):
        super(GeneratorSpherical, self).__init__()
        if r_min < 0 or r_max < r_min:
            raise ValueError(f"Illegal range [{r_min}, {r_max}]")

        if method == 'equally-spaced-noisy':
            lower = r_min ** 2
            upper = r_max ** 2
            rng = upper - lower
            self.get_r = lambda: torch.sqrt(rng * torch.rand(self.shape) + lower)
        elif method == "equally-radius-noisy":
            lower = r_min
            upper = r_max
            rng = upper - lower
            self.get_r = lambda: rng * torch.rand(self.shape) + lower
        else:
            raise ValueError(f'Unknown method: {method}')

        self.size = size  # stored for `solve_spherical_system` to access
        self.r_min = r_min
        self.r_max = r_max
        self.method = method
        self.shape = (size,)  # used for `self.get_example()`

    def get_examples(self):
        a = torch.rand(self.shape)
        b = torch.rand(self.shape)
        c = torch.rand(self.shape)
        denom = a + b + c
        # `x`, `y`, `z` here are just for computation of `theta` and `phi`
        epsilon = 1e-6
        x = torch.sqrt(a / denom) + epsilon
        y = torch.sqrt(b / denom) + epsilon
        z = torch.sqrt(c / denom) + epsilon
        # `sign_x`, `sign_y`, `sign_z` are either -1 or +1
        sign_x = torch.randint(0, 2, self.shape, dtype=x.dtype) * 2 - 1
        sign_y = torch.randint(0, 2, self.shape, dtype=y.dtype) * 2 - 1
        sign_z = torch.randint(0, 2, self.shape, dtype=z.dtype) * 2 - 1

        x = x * sign_x
        y = y * sign_y
        z = z * sign_z

        theta = torch.acos(z).requires_grad_(True)
        phi = -torch.atan2(y, x) + np.pi  # atan2 ranges (-pi, pi] instead of [0, 2pi)
        phi.requires_grad_(True)
        r = self.get_r().requires_grad_(True)

        return r, theta, phi

    def _internal_vars(self) -> dict:
        d = super(GeneratorSpherical, self)._internal_vars()
        d.update(dict(
            r_min=self.r_min,
            r_max=self.r_max,
            method=self.method,
        ))
        return d


class ConcatGenerator(BaseGenerator):
    r"""An concatenated generator for sampling points,
    whose ``get_examples()`` method returns the concatenated vector of the samples returned by its sub-generators.

    :param generators: a sequence of sub-generators, must have a ``.size`` field and a ``.get_examples()`` method
    :type generators: Tuple[BaseGenerator]

    .. note::
        Not to be confused with ``EnsembleGenerator`` which returns all the samples of its sub-generators.
    """

    def __init__(self, *generators):
        super(ConcatGenerator, self).__init__()
        self.generators = generators
        self.size = sum(gen.size for gen in generators)

    def get_examples(self):
        all_examples = [gen.get_examples() for gen in self.generators]
        if isinstance(all_examples[0], torch.Tensor):
            return torch.cat(all_examples)
        # zip(*sequence) is just `unzip`ping a sequence into sub-sequences, refer to this post for more
        # https://stackoverflow.com/questions/19339/transpose-unzip-function-inverse-of-zip
        segmented = zip(*all_examples)
        return [torch.cat(seg) for seg in segmented]

    def _internal_vars(self) -> dict:
        d = super(ConcatGenerator, self)._internal_vars()
        d.update(dict(
            generators=self.generators,
        ))
        return d


class StaticGenerator(BaseGenerator):
    """A generator that returns the same samples every time.
    The static samples are obtained by the sub-generator at instantiation time.

    :param generator: a generator used to generate the static samples
    :type generator: BaseGenerator
    """

    def __init__(self, generator):
        super(StaticGenerator, self).__init__()
        self.generator = generator
        self.size = generator.size
        self.examples = generator.get_examples()

    def get_examples(self):
        return self.examples

    def _internal_vars(self) -> dict:
        d = super(StaticGenerator, self)._internal_vars()
        d.update(dict(
            generator=self.generator,
            examples=self.examples,
        ))
        return d


class PredefinedGenerator(BaseGenerator):
    """A generator for generating points that are fixed and predefined.

    :param xs: The x-dimension of the trianing points
    :type xs: `torch.Tensor`
    :param ys: The y-dimension of the training points
    :type ys: `torch.Tensor`
    """

    def __init__(self, *xs):
        super(PredefinedGenerator, self).__init__()
        self.size = len(xs[0])
        for x in xs:
            if self.size != len(x):
                raise ValueError('tensors of different lengths encountered {self.size} != {len(x)}')
        xs = [x if isinstance(x, torch.Tensor) else torch.tensor(x) for x in xs]
        self.xs = [torch.flatten(x).requires_grad_(True) for x in xs]

        if len(self.xs) == 1:
            self.xs = self.xs[0]

    def get_examples(self):
        """Returns the training points. Points are fixed and predefined.

            :returns: The predefined training points
            :rtype: tuple[`torch.Tensor`]
        """
        return self.xs

    def _internal_vars(self) -> dict:
        d = super(PredefinedGenerator, self)._internal_vars()
        d.update(dict(
            xs=self.xs,
        ))
        return d


class TransformGenerator(BaseGenerator):
    """A generator which applies certain transformations on the sample vectors.

    :param generator:
        A generator used to generate samples on which transformations will be applied.
    :type generator: BaseGenerator
    :param transforms:
        A list of transformations to be applied on the sample vectors.
        Identity transformation can be replaced with None
    :type transforms: list[callable]
    :param transform:
        A callable that transforms the output(s) of base generator to another (tuple of) coordinate(s).
    :type transform: callable
    """

    def __init__(self, generator, transforms=None, transform=None):
        super(TransformGenerator, self).__init__()
        self.generator = generator
        self.size = generator.size
        if transforms is not None and transform is not None:
            raise ValueError("transform and transforms cannot be both specified")
        if transforms is not None:
            self.trans = [
                (lambda x: x) if t is None else t
                for t in transforms
            ]
        elif transform is not None:
            self.trans = transform
        else:
            self.trans = lambda x: x

    def get_examples(self):
        xs = self.generator.get_examples()
        if isinstance(xs, torch.Tensor):
            if callable(self.trans):
                return self.trans(xs)
            else:
                return self.trans[0](xs)
        if callable(self.trans):
            return self.trans(*xs)
        else:
            return tuple(t(x) for t, x in zip(self.trans, xs))

    def _internal_vars(self) -> dict:
        d = super(TransformGenerator, self)._internal_vars()
        d.update(dict(
            generator=self.generator,
            trans=self.trans,
        ))
        return d


class EnsembleGenerator(BaseGenerator):
    r"""A generator for sampling points whose `get_examples` method returns all the samples of its sub-generators.
    All sub-generator must return tensors of the same shape.
    The number of tensors returned by each sub-generator can be different.

    :param generators: a sequence of sub-generators, must have a .size field and a .get_examples() method
    :type generators: Tuple[BaseGenerator]

    .. note::
        Not to be confused with ``ConcatGenerator`` which returns
        the concatenated vector of samples returned by its sub-generators.
    """

    def __init__(self, *generators):
        super(EnsembleGenerator, self).__init__()
        self.size = generators[0].size
        for i, gen in enumerate(generators):
            if gen.size != self.size:
                raise ValueError(f"gens[{i}].size ({gen.size}) != gens[0].size ({self.size})")
        self.generators = generators

    def get_examples(self):
        ret = tuple()
        for g in self.generators:
            ex = g.get_examples()
            if isinstance(ex, list):
                ex = tuple(ex)
            elif isinstance(ex, torch.Tensor):
                ex = (ex,)
            ret += ex

        if len(ret) == 1:
            return ret[0]
        else:
            return ret

    def _internal_vars(self) -> dict:
        d = super(EnsembleGenerator, self)._internal_vars()
        d.update(dict(
            generators=self.generators,
        ))
        return d


class MeshGenerator(BaseGenerator):
    r"""A generator for sampling points whose `get_examples` method returns a mesh of the samples of its sub-generators.
    All sub-generators must return tensors of the same shape, or a tuple of tensors of the same shape.
    The number of tensors returned by each sub-generator can be different, but the intent behind
    this class is to create an N dimensional generator from several 1 dimensional generators, so each input generator
    should represents one of the dimensions of your problem. An exception is made for
    using a ``MeshGenerator`` as one of the inputs of another ``MeshGenerator``. In that case the original
    meshed generators are extracted from the input ``MeshGenerator``, and the final mesh is used using those
    (e.g ``MeshGenerator(MeshGenerator(g1, g2), g3)`` is equivalent to ``MeshGenerator(g1, g2, g3)``, where
    g1, g2 and g3 are ``Generator1D``).
    This is done to make the use of the ^ infix consistent with the use of
    the ``MeshGenerator`` class itself (e.g ``MeshGenerator(g1, g2, g3)`` is equivalent to g1 ^ g2 ^ g3), where
    g1, g2 and g3 are ``Generator1D``).

    :param generators: a sequence of sub-generators, must have a .size field and a .get_examples() method
    :type generators: Tuple[BaseGenerator]
    """

    def __init__(self, *generators):
        super(MeshGenerator, self).__init__()
        self.generators = []
        for g in generators:
            if isinstance(g, MeshGenerator):
                for s in g.generators:
                    self.generators.append(s)
            else:
                self.generators.append(g)
        self.size = np.prod(tuple(g.size for g in self.generators))

    def get_examples(self):
        ret = tuple()
        for g in self.generators:
            ex = g.get_examples()
            if isinstance(ex, list):
                ex = tuple(ex)
            elif isinstance(ex, torch.Tensor):
                ex = (ex,)
            ret += ex

        if len(ret) == 1:
            return ret[0]
        else:
            ret = torch.meshgrid(ret)
            ret_f = tuple()
            for r in ret:
                ret_f += (r.flatten(),)
            return ret_f

    def _internal_vars(self) -> dict:
        d = super(MeshGenerator, self)._internal_vars()
        d.update(dict(
            generators=self.generators,
        ))
        return d


class FilterGenerator(BaseGenerator):
    """A generator which applies some filtering before samples are returned

    :param generator:
        A generator used to generate samples to be filtered.
    :type generator: BaseGenerator
    :param filter_fn:
        A filter to be applied on the sample vectors; maps a list of tensors to a mask tensor.
    :type filter_fn: callable
    :param size:
        Size to be used for `self.size`.
        If not given, this attribute is initialized to the size of ``generator``.
    :type size: int
    :param update_size:
        Whether or not to update `.size` after each call of `self.get_examples`.
        Defaults to True.
    :type update_size: bool
    """

    def __init__(self, generator, filter_fn, size=None, update_size=True):
        super(FilterGenerator, self).__init__()
        self.generator = generator
        self.filter_fn = filter_fn
        if size is None:
            self.size = generator.size
        else:
            self.size = size
        self.update_size = update_size

    def get_examples(self):
        xs = self.generator.get_examples()
        if isinstance(xs, torch.Tensor):
            xs = [xs]
        mask = self.filter_fn(xs)
        xs = [x[mask] for x in xs]
        if self.update_size:
            self.size = len(xs[0])
        if len(xs) == 1:
            return xs[0]
        else:
            return xs

    def _internal_vars(self) -> dict:
        d = super(FilterGenerator, self)._internal_vars()
        d.update(dict(
            generator=self.generator,
            filter_fn=self.filter_fn,
        ))
        return d


class ResampleGenerator(BaseGenerator):
    """A generator whose output is shuffled and resampled every time

    :param generator: A generator used to generate samples to be shuffled and resampled.
    :type generator: BaseGenerator
    :param size: Size of the shuffled output. Defaults to the size of ``generator``.
    :type size: int
    :param replacement: Whether to sample with replacement or not. Defaults to False.
    :type replacement: bool
    """

    def __init__(self, generator, size=None, replacement=False):
        super(ResampleGenerator, self).__init__()
        self.generator = generator
        if size is None:
            self.size = generator.size
        else:
            self.size = size
        self.replacement = replacement

    def get_examples(self):
        if self.replacement:
            indices = torch.randint(self.generator.size, (self.size,))
        else:
            indices = torch.randperm(self.generator.size)[:self.size]

        xs = self.generator.get_examples()
        if isinstance(xs, torch.Tensor):
            return xs[indices]
        else:
            return [x[indices] for x in xs]

    def _internal_vars(self) -> dict:
        d = super(ResampleGenerator, self)._internal_vars()
        d.update(dict(
            generator=self.generator,
            replacement=self.replacement,
        ))
        return d


class BatchGenerator(BaseGenerator):
    """A generator which caches samples and returns a single batch of the samples at a time

    :param generator:
        A generator used for getting (cached) examples.
    :type generator: BaseGenerator
    :param batch_size:
        Number of batches to be returned.
        It can be larger than size of ``generator``, but inefficient if so.
    :type batch_size: int
    """

    def __init__(self, generator, batch_size):
        super(BatchGenerator, self).__init__()

        if generator.size <= 0:
            raise ValueError(f"generator has size {generator.size} <= 0")
        self.generator = generator
        self.size = batch_size
        self.cached_xs = self.generator.get_examples()
        if isinstance(self.cached_xs, torch.Tensor):
            self.cached_xs = [self.cached_xs]
        if isinstance(self.cached_xs, tuple):
            self.cached_xs = list(self.cached_xs)

    def get_examples(self):
        # update cache so that we have enough samples in a batch
        while len(self.cached_xs[0]) < self.size:
            new = self.generator.get_examples()
            if isinstance(new, torch.Tensor):
                new = [new]
            self.cached_xs = [torch.cat([x, n]) for x, n in zip(self.cached_xs, new)]

        batch = [x[:self.size] for x in self.cached_xs]
        # drop the returned samples
        self.cached_xs = [x[self.size:] for x in self.cached_xs]

        if len(batch) == 1:
            return batch[0]
        else:
            return batch

    def _internal_vars(self) -> dict:
        d = super(BatchGenerator, self)._internal_vars()
        d.update(dict(
            generator=self.generator,
        ))
        return d


class SamplerGenerator(BaseGenerator):
    def __init__(self, generator):
        super(SamplerGenerator, self).__init__()
        self.generator = generator
        self.size = generator.size

    def get_examples(self) -> List[torch.Tensor]:
        samples = self.generator.get_examples()
        if isinstance(samples, torch.Tensor):
            samples = [samples]
        samples = [u.reshape(-1, 1) for u in samples]
        return samples

    def _internal_vars(self) -> dict:
        d = super(SamplerGenerator, self)._internal_vars()
        d.update(dict(
            generator=self.generator,
        ))
        return d



import os
import json
import dill
import numpy as np
import pathlib
import torch
import requests
import tempfile
from typing import Union
from itertools import chain
import inspect
import ast
import types

# Is Dev mode
try:
    DEV = os.environ["DEV"]
    DEV = True
except KeyError:
    DEV = False

try:
    NEURODIFF_API_URL = os.environ["NEURODIFF_API_URL"]
except KeyError:
    if DEV:
        NEURODIFF_API_URL = "http://dev.neurodiff.io/api/v1"
    else:
        NEURODIFF_API_URL = "http://www.neurodiff.io/api/v1"


def is_solution_name(name):
    if name.startswith('./'):
        return False
    else:
        return True


def process_response(response):
    """
    Process a `requests.Reponse` object, returning the decoded contents or
    raising an APIError (or subclass) exception on request failure.
    """
    return response.json()


def _make_api_headers():
    headers = {}
    try:
        NEURODIFF_API_KEY = os.environ["NEURODIFF_API_KEY"]
    except KeyError:
        print("No API Key was found in environment variable NEURODIFF_API_KEY")
        NEURODIFF_API_KEY = ""

    headers["apikey"] = NEURODIFF_API_KEY

    return headers


def create_cache_dir():
    cache_dir = os.path.join(os.path.expanduser('~'), '.neurodiff')
    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)

    return cache_dir


def get_file(url, name):
    cache_dir = create_cache_dir()
    solution_file_path = os.path.join(cache_dir, name.replace("/", "_"))

    # if not os.path.exists(solution_file_path):
    url = url + "?name="+name
    # Download the solution
    with requests.get(url, stream=True, headers=_make_api_headers()) as r:
        r.raise_for_status()
        with open(solution_file_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return solution_file_path


def get_source(lambda_function):
    lambda_text = ""
    try:
        source_lines, _ = inspect.getsourcelines(lambda_function)
        lambda_text = "".join([line.strip() for line in source_lines])
        source_ast = ast.parse(lambda_text)
        lambda_node = next((node for node in ast.walk(source_ast)
                            if isinstance(node, ast.Lambda)), None)
        lambda_text = lambda_text[lambda_node.col_offset:]
    except:
        pass

    return lambda_text


def get_parameters(lambda_function):
    parameters = {}
    try:
        closures = lambda_function.__closure__
        if closures is not None:
            freevars = lambda_function.__code__.co_freevars
            for i, c in enumerate(closures):
                parameters[freevars[i]] = c.cell_contents
        else:
            gbs = lambda_function.__globals__
            co_names = lambda_function.__code__.co_names
            for i, c in enumerate(co_names):
                if c != "diff" and c != "torch":
                    parameters[c] = gbs[c]
    except:
        pass

    return parameters


def get_conditions(conditions):
    condition_list = []
    try:
        for condition in conditions:
            cond_dict = condition.__dict__
            cond_dict["condition_type"] = condition.__class__.__name__

            # Get the lambda functions in the condition
            for key, value in cond_dict.items():
                if isinstance(value, types.FunctionType):
                    function_source = get_source(value)
                    if function_source != "":
                        cond_dict[key] = function_source

            condition_list.append(cond_dict)
    except:
        pass
    return condition_list


def get_generator(generator):
    gen_dict = {}
    try:
        gen_dict = generator["train"].__dict__['generator'].__dict__.copy()
        if "examples" in gen_dict:
            del gen_dict['examples']

        if "grid_x" in gen_dict:
            del gen_dict['grid_x']
        if "grid_y" in gen_dict:
            del gen_dict['grid_y']
        del gen_dict['getter']
    except:
        pass
    return gen_dict


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(JsonEncoder, self).default(obj)


def get_sample_solution1D(solver):
    sample_solution_curve = []
    try:
        t = np.linspace(solver.t_min, solver.t_max, 10 *
                        (int(solver.t_max-solver.t_min)))
        sample_solution = solver.get_solution()(t)

        if not isinstance(sample_solution, list):
            sample_solution = [sample_solution]

        for i in range(len(sample_solution)):
            sample_solution[i] = sample_solution[i].cpu(
            ).detach().numpy().tolist()

        sample_solution_curve = [t.tolist(), sample_solution]
    except:
        pass
    return sample_solution_curve


def get_sample_solution2D(solver):
    sample_solution_curve = []

    try:
        inputs = solver.generator['train'].get_examples()
        sample_solution = solver.get_solution()(
            inputs[0].view(-1), inputs[1].view(-1))
        sample_solution = sample_solution.cpu().detach().numpy()
        for i in range(2):
            inputs[i] = inputs[i].view(-1).cpu().detach().numpy().tolist()
        sample_solution_curve = [inputs, sample_solution.reshape(
            solver.generator["train"].__dict__['generator'].__dict__['grid']).tolist()]
    except:
        pass
    return sample_solution_curve


def get_networks(solver):
    networks = []
    for idx, net in enumerate(solver.nets):
        layers = []
        for layer in net.NN:
            layer_params = {
                "layer": layer.__class__.__name__
            }
            layer_dict = layer.__dict__
            if "in_features" in layer_dict:
                layer_params["in_features"] = layer_dict["in_features"]
            if "out_features" in layer_dict:
                layer_params["out_features"] = layer_dict["out_features"]
            if "bias" in layer_dict:
                layer_params["bias"] = layer_dict["bias"]

            layers.append(layer_params)
        networks.append({"layers": layers})

    return networks


def get_loss(loss):
    try:
        ...
#        loss_sample = json.dumps(loss,cls = JsonEncoder)
#        loss_sample = loss
    except:
        pass
    return loss_sample


class SolverConfig():
    conditions = None
    ode_system = None
    pde_system = None
    nets = None
    optimizer = None
    optimizer_params = None
    train_generator = None
    valid_generator = None


class PretrainedSolver():
    diff_eqs_source = ""

    def print_diff_eqs(self):
        lambda_text = get_source(self.diff_eqs)
        if lambda_text == "":
            lambda_text = self.diff_eqs_source
        print(lambda_text)

    # Saving selected attributes of model in dict

    def save(self,
             path: str = None,
             name: str = None,
             save_to_hub=False):

        # Check params
        if path is None and save_to_hub == False:
            raise Exception("path cannot be empty when save_to_hub=False")
        if name is None and save_to_hub == True:
            raise Exception("name cannot be empty when save_to_hub=True")

        # Check if optimizer is existing in pytorch
        optimizer_class = None
        for cls in torch.optim.Optimizer.__subclasses__():
            if self.optimizer.__class__.__name__ == cls.__name__:
                optimizer_class = self.optimizer.__class__

        # Get Diff equations details
        if self.__class__.__name__ == "Solver1D":
            sample_solution = get_sample_solution1D(self)
        elif self.__class__.__name__ == "Solver2D":
            sample_solution = get_sample_solution2D(self)

        diff_equation_details = {
            "equation": get_source(self.diff_eqs),
            "parameters": get_parameters(self.diff_eqs),
            "conditions": get_conditions(self.conditions),
            "generator": get_generator(self.generator),
            "sample_solution": sample_solution,
            "sample_loss": self.metrics_history['valid_loss'],
            "criterion": get_source(self.criterion),
            "networks": get_networks(self),
            "optimizer": {
                "name": self.optimizer.__class__.__name__,
                "params": self.optimizer.state_dict()['param_groups']
            },
        }

        save_dict = {
            "metrics": self.metrics_fn,
            "criterion": self.criterion,
            "conditions": self.conditions,
            "global_epoch": self.global_epoch,  # loss_history
            "nets": self.nets,
            "optimizer": self.optimizer,
            "optimizer_state": self.optimizer.state_dict(),
            "optimizer_class": optimizer_class,
            "diff_eqs": self.diff_eqs,
            "diff_equation_details": diff_equation_details,
            "generator": self.generator,
            "train_loss_history": self.metrics_history['train_loss'],
            "valid_loss_history": self.metrics_history['valid_loss'],
            "type": self.__class__,
            "type_name": self.__class__.__name__,
            "solver": self
        }

        # Save remote if needed
        if save_to_hub:
            cache_dir = create_cache_dir()
            solution_file_path = os.path.join(cache_dir, "solution_to_upload")
            # Save solution in temp file
            # with tempfile.NamedTemporaryFile() as tmp_file:
            with open(solution_file_path, 'wb') as tmp_file:
                dill.dump(save_dict, tmp_file)
                # tmp_file.flush()

                # Save remote
                print("Saving solution to:", NEURODIFF_API_URL)
                project = None
                if "/" in name:
                    project = name.split('/')[0]
                if project is None:
                    print("Default project will be used to save solution")
                else:
                    # Check if user has access to project
                    url = NEURODIFF_API_URL + \
                        "/projects/check_access/{project}"
                    response = requests.get(
                        url.format(project=project),
                        headers=_make_api_headers()
                    )
                    if not response.ok:
                        # response.raise_for_status()
                        print("You do not have access to the project:", project)

                # Upload the solution
                url = NEURODIFF_API_URL + "/solutions/upload"
                solution = {
                    "name": name,
                    "description": name,
                    "diff_equation_details": save_dict["diff_equation_details"],
                    "type_name": save_dict["type_name"]
                }
                response = requests.post(
                    url,
                    data=solution,
                    files={"file": open(tmp_file.name, "rb"), "solution": (
                        "solution.json", json.dumps(solution))},
                    headers=_make_api_headers()
                )
                if not response.ok:
                    print("Could not upload solution")

        else:
            # Save solution locally
            with open(path, 'wb') as file:
                dill.dump(save_dict, file)

    # Loading saved attributes into new solver object

    # Have to add check and warning/error

    @classmethod
    def load(cls,
             path: str = None,
             name: str = None,
             config=SolverConfig()):

        # Check params
        if path is None and name is None:
            raise Exception("Either path or name is required to load solver")

        # Load from remote
        if path is None:
            print("Loading solution from:", NEURODIFF_API_URL, name)
            url = NEURODIFF_API_URL + "/solutions/download"
            solution_file_path = get_file(url, name)
        else:
            print("Loading solution from:", path)
            solution_file_path = path
        # Load the solution
        with open(solution_file_path, 'rb') as file:
            load_dict = dill.load(file)

        # Loading user defined generator and extracting time domain information
        if config.train_generator == None:
            train_generator = load_dict['generator']['train'].generator
            valid_generator = load_dict['generator']['valid'].generator
        else:
            train_generator = config.train_generator
            valid_generator = config.valid_generator

        # Loading user defined ode_system or system from load file
        if (config.ode_system == None) and (config.pde_system == None):
            de_system = load_dict['diff_eqs']
        elif config.ode_system is not None:
            de_system = config.ode_system
        elif config.pde_system is not None:
            de_system = config.pde_system

        # Loading user defined conditions or conditions from load file
        if config.conditions == None:
            cond = load_dict['conditions']
        else:
            cond = config.conditions

        # Loading user defined nets or nets from load file
        if config.nets == None:
            nets = load_dict['nets']
        else:
            nets = config.nets

        # Loading user defined optimizer or optimizer from load file

        train_loss = []
        valid_loss = []
        if config.optimizer == None:
            #optimizer = load_dict['optimizer']
            if load_dict['optimizer_class'] is not None:
                optimizer = load_dict['optimizer_class'](
                    chain.from_iterable(n.parameters() for n in nets))
                optimizer.load_state_dict(load_dict['optimizer_state'])
            else:
                optimizer = load_dict['optimizer']

            # As older network/optimizer is loaded load loss history as well
            train_loss = load_dict['train_loss_history']
            valid_loss = load_dict['valid_loss_history']
        else:
            # Declare a flag to check if optimizer is passed as a class. Link parameters and load state to the new class
            check_flag = False

            if config.optimizer_params is not None:
                optimizer = config.optimizer(chain.from_iterable(
                    n.parameters() for n in nets), **config.optimizer_params)
                train_loss = load_dict['train_loss_history']
                valid_loss = load_dict['valid_loss_history']
                check_flag = True

            # In case of custom optimizer or optimizer externally linked to the network, load complete optimizer
            if check_flag == False:
                optimizer = load_dict['optimizer']

        # Initiate a new Solver
        if load_dict["type_name"] == "Solver1D":
            # t min/max
            t_min = load_dict['generator']['train'].__dict__[
                'generator'].__dict__['t_min']
            t_max = load_dict['generator']['train'].__dict__[
                'generator'].__dict__['t_max']

            solver = cls(ode_system=de_system,
                         conditions=cond,
                         criterion=load_dict['criterion'],
                         metrics=load_dict['metrics'],
                         nets=nets,
                         optimizer=optimizer,
                         train_generator=train_generator,
                         valid_generator=valid_generator,
                         t_min=t_min,
                         t_max=t_max)
        elif load_dict["type_name"] == "Solver2D":
            xy_min = load_dict['generator']['train'].__dict__[
                'generator'].__dict__['xy_min']
            xy_max = load_dict['generator']['train'].__dict__[
                'generator'].__dict__['xy_max']

            solver = cls(pde_system=de_system,
                         conditions=cond,
                         xy_min=xy_min,
                         xy_max=xy_max,
                         nets=nets,
                         train_generator=train_generator,
                         valid_generator=valid_generator,
                         optimizer=optimizer,
                         criterion=load_dict['criterion'],
                         metrics=load_dict['metrics'])

        solver.metrics_history['train_loss'] = train_loss
        solver.metrics_history['valid_loss'] = valid_loss

        try:
            solver.diff_eqs_source = load_dict["diff_equation_details"]["equation"]
        except:
            pass
        return solver



import warnings
import functools


def warn_deprecate_class(new_class):
    """get a new class name that issues a warning when instantiated
    :param new_class: new class name
    :type new_class: type
    :return: a function that, when called, acts as if it is a class constructor
    :rtype: callable
    """
    @functools.wraps(new_class)
    def old_class_getter(*args, **kwargs):
        warnings.warn(f"This class name is deprecated, use {new_class} instead", FutureWarning)
        return new_class(*args, **kwargs)

    return old_class_getter


def deprecated_alias(**aliases):
    """A decorator to deprecate old argument names in favor of new ones.
    See more here https://stackoverflow.com/a/49802489.

    :param aliases: A sequence of keyword argument of the form: old_name="name_name"
    :param aliases: Dict[str,str]
    :return: A decorated function that can receive either `old_name` or `new_name` as input
    :rtype: function
    """
    def deco(f):
        @functools.wraps(f)  # preserves signature and docstring
        def wrapper(*args, **kwargs):
            _rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)
        return wrapper
    return deco


def _rename_kwargs(func_name, kwargs, aliases):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise KeyError(f'{func_name} received both `{alias}` (deprecated) and `{new}` (recommended)')
            warnings.warn(f'The argument `{alias}` is deprecated; use `{new}` instead for {func_name}.', FutureWarning)
            kwargs[new] = kwargs.pop(alias)